import importlib
import inspect
import logging
import os
import shutil
import sys
import tempfile
import warnings
from copy import deepcopy
from http.cookiejar import LoadError
from typing import Union

import cloudpickle
import numpy as np
import torch
import torch.distributions as dist
import torch.nn as nn
from pythae.models.base.base_utils import CPU_Unpickler, ModelOutput
from pythae.models.nn.base_architectures import BaseDecoder, BaseEncoder
from torch.nn import functional as F

from ...data.datasets.base import MultimodalBaseDataset
from ..auto_model import AutoConfig
from ..nn.default_architectures import BaseDictDecoders, BaseDictEncoders
from .base_config import BaseMultiVAEConfig, EnvironmentConfig
from .base_utils import hf_hub_is_available, model_card_template

logger = logging.getLogger(__name__)
console = logging.StreamHandler()
logger.addHandler(console)
logger.setLevel(logging.INFO)


def cross_entropy(input, target, eps=1e-6):
    """k-Class Cross Entropy (Log Softmax + Log Loss)

    @param input: torch.Tensor (size K x N x d)
    @param target: torch.Tensor (size N x d) or (d,)
    @param eps: error to add (default: 1e-6)
    @return loss: torch.Tensor (size N)
    """
    if isinstance(target, dict):
        # converts to tokens proba instead of class id for text
        _target = torch.zeros(input.shape[0] * input.shape[1], input.shape[-1]).to(
            input.device
        )
        _target = _target.scatter(1, target["tokens"].reshape(-1, 1), 1)
        input = input.reshape(input.shape[0] * input.shape[1], -1)
    else:
        _target = target

    log_input = F.log_softmax(input + eps, dim=-1)
    loss = _target * log_input
    return loss


class BaseMultiVAE(nn.Module):
    """Base class for Multimodal VAE models.

    Args:
        model_config (BaseMultiVAEConfig): An instance of BaseMultiVAEConfig in which any model's
            parameters is made available.

        encoders (Dict[str, ~pythae.models.nn.base_architectures.BaseEncoder]): A dictionary containing
            the modalities names and the encoders for each modality. Each encoder is an instance of
            Pythae's BaseEncoder. Default: None.

        decoder (Dict[str, ~pythae.models.nn.base_architectures.BaseDecoder]): A dictionary containing
            the modalities names and the decoders for each modality. Each decoder is an instance of
            Pythae's BaseDecoder.


    """

    def __init__(
        self,
        model_config: BaseMultiVAEConfig,
        encoders: dict = None,
        decoders: dict = None,
    ):
        nn.Module.__init__(self)

        self.model_name = "BaseMultiVAE"
        self.model_config = model_config
        self.n_modalities = model_config.n_modalities
        self.input_dims = model_config.input_dims
        self.reset_optimizer_epochs = []
        self.multiple_latent_spaces = False  # Default value, this field must be changes
        # in models using multiple latent spaces

        if encoders is None:
            if self.input_dims is None:
                raise AttributeError(
                    "Please provide encoders or input dims for the modalities in the model_config."
                )
            else:
                if len(self.input_dims.keys()) != self.n_modalities:
                    raise AttributeError(
                        f"The provided number of input_dims {len(self.input_dims.keys())} doesn't"
                        f"match the number of modalities ({self.n_modalities} in model config "
                    )
                encoders = self.default_encoders(model_config)
        else:
            self.model_config.custom_architectures.append("encoders")

        if decoders is None:
            if self.input_dims is None:
                raise AttributeError(
                    "Please provide decoders or input dims for the modalities in the model_config."
                )
            else:
                if len(self.input_dims.keys()) != self.n_modalities:
                    raise AttributeError(
                        f"The provided number of input_dims {len(self.input_dims.keys())} doesn't"
                        f"match the number of modalities ({self.n_modalities} in model config "
                    )
                decoders = self.default_decoders(model_config)
        else:
            self.model_config.custom_architectures.append("decoders")

        self.sanity_check(encoders, decoders)

        self.latent_dim = model_config.latent_dim
        self.model_config = model_config
        self.device = None

        self.set_decoders(decoders)
        self.set_encoders(encoders)

        self.modalities_name = list(self.decoders.keys())

        # Check that the modalities' name are coherent
        if self.input_dims is not None:
            if self.input_dims.keys() != self.encoders.keys():
                raise KeyError(
                    f"Warning! : The modalities names in model_config.input_dims : {list(self.input_dims.keys())}"
                    f" does not match the modalities names in encoders : {list(self.encoders.keys())}"
                )

        self.use_likelihood_rescaling = model_config.uses_likelihood_rescaling
        if self.use_likelihood_rescaling:
            if self.model_config.rescale_factors is not None:
                self.rescale_factors = model_config.rescale_factors
            elif self.input_dims is None:
                raise AttributeError(
                    " inputs_dim = None but (use_likelihood_rescaling = True"
                    " in model_config)"
                    " To compute likelihood rescalings we need the input dimensions."
                    " Please provide a valid dictionary for input_dims."
                )
            else:
                max_dim = max(*[np.prod(self.input_dims[k]) for k in self.input_dims])
                self.rescale_factors = {
                    k: max_dim / np.prod(self.input_dims[k]) for k in self.input_dims
                }
        else:
            self.rescale_factors = {k: 1 for k in self.encoders}
            # above, we take the modalities keys in self.encoders as input_dims may be None

        # Set the reconstruction losses
        if model_config.decoders_dist is None:
            model_config.decoders_dist = {k: "normal" for k in self.encoders}
        if model_config.decoder_dist_params is None:
            model_config.decoder_dist_params = {}
        self.set_decoders_dist(
            model_config.decoders_dist, deepcopy(model_config.decoder_dist_params)
        )

    def set_decoders_dist(self, recon_dict, dist_params_dict):
        """Set the reconstruction losses functions decoders_dist
        and the log_probabilites functions recon_log_probs.
        recon_log_probs is the normalized negative version of recon_loss and is used only for
        likelihood estimation.
        """
        self.recon_log_probs = {}

        for k in recon_dict:
            if recon_dict[k] == "normal":
                params_mod = dist_params_dict.pop(k, {})
                scale = params_mod.pop("scale", 1.0)
                self.recon_log_probs[k] = lambda input, target: dist.Normal(
                    input, scale
                ).log_prob(target)

            elif recon_dict[k] == "bernoulli":
                self.recon_log_probs[k] = lambda input, target: dist.Bernoulli(
                    logits=input
                ).log_prob(target)

            elif recon_dict[k] == "laplace":
                params_mod = dist_params_dict.pop(k, {})
                scale = params_mod.pop("scale", 1.0)
                self.recon_log_probs[k] = lambda input, target: dist.Laplace(
                    input, scale
                ).log_prob(target)

            elif recon_dict[k] == "categorical":
                self.recon_log_probs[k] = lambda input, target: cross_entropy(
                    input, target
                )

        # TODO : add the possibility to provide custom reconstruction loss and in that case use the negative
        # reconstruction loss as the log probability.

    def sanity_check(self, encoders, decoders):
        if self.n_modalities != len(encoders.keys()):
            raise AttributeError(
                f"The provided number of encoders {len(encoders.keys())} doesn't"
                f"match the number of modalities ({self.n_modalities} in model config "
            )

        if self.n_modalities != len(decoders.keys()):
            raise AttributeError(
                f"The provided number of decoders {len(decoders.keys())} doesn't"
                f"match the number of modalities ({self.n_modalities} in model config "
            )

        if encoders.keys() != decoders.keys():
            raise AttributeError(
                "The names of the modalities in the encoders dict doesn't match the names of the modalities"
                " in the decoders dict."
            )

    def encode(
        self,
        inputs: MultimodalBaseDataset,
        cond_mod: Union[list, str] = "all",
        N: int = 1,
        **kwargs,
    ) -> ModelOutput:
        """
        Generate encodings conditioning on all modalities or a subset of modalities.

        Args:
            inputs (MultimodalBaseDataset): The dataset to use for the conditional generation.
            cond_mod (Union[list, str]): Either 'all' or a list of str containing the modalities names to condition on.
            N (int) : The number of encodings to sample for each datapoint. Default to 1.

        """

        # If the input cond_mod is a string : convert it to a list
        if type(cond_mod) == str:
            if cond_mod == "all":
                cond_mod = list(self.encoders.keys())
            elif cond_mod in self.encoders.keys():
                cond_mod = [cond_mod]
            else:
                raise AttributeError(
                    'If cond_mod is a string, it must either be "all" or a modality name'
                    f" The provided string {cond_mod} is neither."
                )

        ignore_incomplete = kwargs.pop("ignore_incomplete", False)
        # Deal with incomplete datasets
        if hasattr(inputs, "masks") and not ignore_incomplete:
            # Check that all modalities in cond_mod are available for all samples points.
            mods_avail = torch.tensor(True)
            for m in cond_mod:
                mods_avail = torch.logical_and(mods_avail, inputs.masks[m])
            if not torch.all(mods_avail):
                raise AttributeError(
                    "You tried to encode a incomplete dataset conditioning on",
                    f"modalities {cond_mod}, but some samples are not available"
                    "in all those modalities.",
                )
        return ModelOutput(cond_mod=cond_mod)

    def decode(self, embedding: ModelOutput, modalities: Union[list, str] = "all"):
        """Decode a latent variable z in all modalities specified in modalities.

        Args:
            z (ModelOutput): the latent variables. In case there is only one latent space, z is a tensor
                otherwise it is a dictionary containing all the latent variables associated with modalities'name.
            modalities (Union(List, str), Optional): the modalities to decode from z. Default to 'all'.
        Return
            ModelOutput : containing a tensor per modality name.
        """
        self.eval()
        if modalities == "all":
            modalities = list(self.decoders.keys())
        elif type(modalities) == str:
            modalities = [modalities]

        if embedding.one_latent_space:
            z = embedding.z
            outputs = ModelOutput()
            for m in modalities:
                outputs[m] = self.decoders[m](z).reconstruction
            return outputs
        else:
            z_content = embedding.z
            outputs = ModelOutput()

            for m in modalities:
                z = torch.cat([z_content, embedding.modalities_z[m]], dim=-1)
                outputs[m] = self.decoders[m](z).reconstruction
            return outputs

    def predict(
        self,
        inputs: MultimodalBaseDataset,
        cond_mod: Union[list, str] = "all",
        gen_mod: Union[list, str] = "all",
        N: int = 1,
        flatten: bool = False,
        **kwargs,
    ):
        """Generate in all modalities conditioning on a subset of modalities.

        Args:
            inputs (MultimodalBaseDataset): The data to condition on. It must contain at least the modalities
                contained in cond_mod.
            cond_mod (Union[list, str], optional): The modalities to condition on. Defaults to 'all'.
            gen_mod (Union[list, str], optional): The modalities to generate. Defaults to 'all'.
            N (int) : Number of samples to generate. Default to 1.
            flatten (int) : If N>1 and flatten is False, the returned samples have dimensions (N,len(inputs),...).
                Otherwise, the returned samples have dimensions (len(inputs)*N, ...)

        """
        self.eval()
        ignore_incomplete = kwargs.pop("ignore_incomplete", False)
        z = self.encode(
            inputs,
            cond_mod,
            N=N,
            flatten=True,
            ignore_incomplete=ignore_incomplete,
            **kwargs,
        )
        output = self.decode(z, gen_mod)
        n_data = len(z.z) // N
        if not flatten and N > 1:
            for m in self.encoders:
                output[m] = output[m].reshape(N, n_data, *output[m].shape[1:])
        return output

    def forward(self, inputs: MultimodalBaseDataset, **kwargs) -> ModelOutput:
        """
        Main forward pass outputing the VAE outputs
        This function should output a :class:`~pythae.models.base.base_utils.ModelOutput` instance
        gathering all the model outputs

        Args:
            inputs (BaseDataset): The training data with labels, masks etc...

        Returns:
            ModelOutput: A ModelOutput instance providing the outputs of the model.

        .. note::
            The loss must be computed in this forward pass and accessed through
            ``loss = model_output.loss``
        """
        raise NotImplementedError()

    def update(self):
        """Method that allows model update during the training (at the end of a training epoch)

        If needed, this method must be implemented in a child class.

        By default, it does nothing.
        """
        pass

    def default_encoders(self, model_config) -> nn.ModuleDict:
        return BaseDictEncoders(self.input_dims, model_config.latent_dim)

    def default_decoders(self, model_config) -> nn.ModuleDict:
        return BaseDictDecoders(self.input_dims, model_config.latent_dim)

    def set_encoders(self, encoders: dict) -> None:
        """Set the encoders of the model"""
        self.encoders = nn.ModuleDict()
        for modality in encoders:
            encoder = encoders[modality]
            if not issubclass(type(encoder), BaseEncoder):
                raise AttributeError(
                    (
                        f"For modality {modality}, encoder must inherit from BaseEncoder class from "
                        "pythae.models.base_architectures.BaseEncoder. Refer to documentation."
                    )
                )

            self.encoders[modality] = encoder

    def set_decoders(self, decoders: dict) -> None:
        """Set the decoders of the model"""
        self.decoders = nn.ModuleDict()
        for modality in decoders:
            decoder = decoders[modality]
            if not issubclass(type(decoder), BaseDecoder):
                raise AttributeError(
                    (
                        f"For modality {modality}, decoder must inherit from BaseDecoder class from "
                        "pythae.models.base_architectures.BaseDecoder. Refer to documentation."
                    )
                )
            self.decoders[modality] = decoder

    def save(self, dir_path: str):
        """Method to save the model at a specific location. It saves, the model weights as a
        ``models.pt`` file along with the model config as a ``model_config.json`` file. If the
        model to save used custom encoder (resp. decoder) provided by the user, these are also
        saved as ``encoders.pkl`` (resp. ``decoders.pkl``).

        Args:
            dir_path (str): The path where the model should be saved. If the path
                path does not exist a folder will be created at the provided location.
        """
        env_spec = EnvironmentConfig(
            python_version=f"{sys.version_info[0]}.{sys.version_info[1]}"
        )

        model_dict = {"model_state_dict": deepcopy(self.state_dict())}

        if not os.path.exists(dir_path):
            try:
                os.makedirs(dir_path)

            except FileNotFoundError as e:
                raise e

        env_spec.save_json(dir_path, "environment")
        self.model_config.save_json(dir_path, "model_config")

        for archi in self.model_config.custom_architectures:
            with open(os.path.join(dir_path, archi + ".pkl"), "wb") as fp:
                cloudpickle.register_pickle_by_value(
                    inspect.getmodule(self.__getattr__(archi))
                )
                cloudpickle.dump(self.__getattr__(archi), fp)

        torch.save(model_dict, os.path.join(dir_path, "model.pt"))

    @classmethod
    def _load_model_config_from_folder(cls, dir_path):
        file_list = os.listdir(dir_path)

        if "model_config.json" not in file_list:
            raise FileNotFoundError(
                f"Missing model config file ('model_config.json') in"
                f"{dir_path}... Cannot perform model building."
            )

        path_to_model_config = os.path.join(dir_path, "model_config.json")
        model_config = AutoConfig.from_json_file(path_to_model_config)

        return model_config

    @classmethod
    def _load_model_weights_from_folder(cls, dir_path):
        file_list = os.listdir(dir_path)

        if "model.pt" not in file_list:
            raise FileNotFoundError(
                f"Missing model weights file ('model.pt') file in"
                f"{dir_path}... Cannot perform model building."
            )

        path_to_model_weights = os.path.join(dir_path, "model.pt")

        try:
            model_weights = torch.load(path_to_model_weights, map_location="cpu")

        except RuntimeError:
            RuntimeError(
                "Enable to load model weights. Ensure they are saves in a '.pt' format."
            )

        if "model_state_dict" not in model_weights.keys():
            raise KeyError(
                "Model state dict is not available in 'model.pt' file. Got keys:"
                f"{model_weights.keys()}"
            )

        model_weights = model_weights["model_state_dict"]

        return model_weights

    @classmethod
    def _load_custom_archi_from_folder(cls, dir_path, archi: str):
        file_list = os.listdir(dir_path)
        cls._check_python_version_from_folder(dir_path=dir_path)

        if archi + ".pkl" not in file_list:
            raise FileNotFoundError(
                f"Missing architecture pkl file ('{archi}.pkl') in"
                f"{dir_path}... This file is needed to rebuild custom encoders."
                " Cannot perform model building."
            )

        else:
            with open(os.path.join(dir_path, f"{archi}.pkl"), "rb") as fp:
                archi = CPU_Unpickler(fp).load()

        return archi

    @classmethod
    def load_from_folder(cls, dir_path: str):
        """Class method to be used to load the model from a specific folder

        Args:
            dir_path (str): The path where the model should have been be saved.

        .. note::
            This function requires the folder to contain:

            - | a ``model_config.json`` and a ``model.pt`` if no custom architectures were provided

            **or**

            - | a ``model_config.json``, a ``model.pt`` and a ``encoders.pkl`` (resp.
                ``decoders.pkl``) if a custom encoders (resp. decoders) were provided
        """

        model_config = cls._load_model_config_from_folder(dir_path)
        model_weights = cls._load_model_weights_from_folder(dir_path)

        custom_architectures = {}
        for archi in model_config.custom_architectures:
            custom_architectures[archi] = cls._load_custom_archi_from_folder(
                dir_path, archi
            )

        model = cls(model_config, **custom_architectures)
        model.load_state_dict(model_weights)

        return model

    @classmethod
    def _check_python_version_from_folder(cls, dir_path: str):
        if "environment.json" in os.listdir(dir_path):
            env_spec = EnvironmentConfig.from_json_file(
                os.path.join(dir_path, "environment.json")
            )
            python_version = env_spec.python_version
            python_version_minor = python_version.split(".")[1]

            if python_version_minor == "8" and sys.version_info[1] > 8:
                raise LoadError(
                    "Trying to reload a model saved with python3.8 with python3.9+. "
                    "Please create a virtual env with python 3.8 to reload this model."
                )

            elif int(python_version_minor) >= 9 and sys.version_info[1] == 8:
                raise LoadError(
                    "Trying to reload a model saved with python3.9+ with python3.8. "
                    "Please create a virtual env with python 3.9+ to reload this model."
                )

    def compute_joint_nll(
        self, inputs: MultimodalBaseDataset, K: int = 1000, batch_size_K: int = 100
    ):
        raise NotImplementedError

    def compute_cond_nll(
        self,
        inputs: MultimodalBaseDataset,
        cond_mod: str,
        pred_mods: list,
        K: int = 1000,
        batch_size_K: int = 100,
    ):
        """Compute the conditional likelihoods ln p(x|y) , ln p(y|x) with MonteCarlo Sampling and the approximation :

                ln p(x|y) = \sum_{z ~ q(z|y)} ln p(x|z)

        Args:
            inputs (MultimodalBaseDataset): the data to compute the likelihood on.
            cond_mod (str): the modality to condition on
            gen_mod (str): the modality to condition on
            K (int, optional): number of samples per batch. Defaults to 1000.
            batch_size_K (int, optional): _description_. Defaults to 100.

        Returns:
            dict: _description_
        """

        # Compute K samples for each datapoint
        o = self.encode(inputs, cond_mod, N=K)

        # Compute the negative recon_log_prob for each datapoint
        ll = {k: [] for k in pred_mods}

        n_data = len(inputs.data[list(inputs.data.keys())[0]])
        for i in range(n_data):
            start_idx, stop_index = 0, batch_size_K
            lnpxs = {k: [] for k in pred_mods}

            while stop_index <= K:
                # Encode with the conditional VAE
                latents = o.z[start_idx:stop_index]

                # Decode with the opposite decoder
                for k in pred_mods:
                    target = inputs.data[k][i]
                    recon = self.decoders[k](
                        latents
                    ).reconstruction  # (batch_size,*recon_shape)
                    # Compute lnp(y|z)
                    lpxz = (
                        self.recon_log_probs[k](target, recon)
                        .reshape(recon.size(0), -1)
                        .sum(-1)
                    )
                    lnpxs[k].append(torch.logsumexp(lpxz, dim=0))

                # next batch
                start_idx += batch_size_K
                stop_index += batch_size_K
            for k in pred_mods:
                ll[k].append(torch.logsumexp(torch.tensor(lnpxs[k]), dim=0) - np.log(K))

        results = {}
        for k in pred_mods:
            results["ll_" + cond_mod + "_" + k] = torch.sum(torch.tensor(ll[k])) / len(
                ll[k]
            )

        return ModelOutput(**results)

    def push_to_hf_hub(self, hf_hub_path: str):  # pragma: no cover
        # TODO : add the load from hf hub function in AutoModel

        """Method allowing to save your model directly on the huggung face hub.
        You will need to have the `huggingface_hub` package installed and a valid Hugging Face
        account. You can install the package using

        .. code-block:: bash

            python -m pip install huggingface_hub

        end then login using

        .. code-block:: bash

            huggingface-cli login

        Args:
            hf_hub_path (str): path to your repo on the Hugging Face hub.
        """
        if not hf_hub_is_available():
            raise ModuleNotFoundError(
                "`huggingface_hub` package must be installed to push your model to the HF hub. "
                "Run `python -m pip install huggingface_hub` and log in to your account with "
                "`huggingface-cli login`."
            )

        else:
            from huggingface_hub import CommitOperationAdd, HfApi

        logger.info(
            f"Uploading {self.model_name} model to {hf_hub_path} repo in HF hub..."
        )

        tempdir = tempfile.mkdtemp()

        self.save(tempdir)

        model_files = os.listdir(tempdir)

        api = HfApi()
        hf_operations = []

        for file in model_files:
            hf_operations.append(
                CommitOperationAdd(
                    path_in_repo=file,
                    path_or_fileobj=f"{str(os.path.join(tempdir, file))}",
                )
            )

        with open(os.path.join(tempdir, "model_card.md"), "w") as f:
            f.write(model_card_template)

        hf_operations.append(
            CommitOperationAdd(
                path_in_repo="README.md",
                path_or_fileobj=os.path.join(tempdir, "model_card.md"),
            )
        )

        try:
            api.create_commit(
                commit_message=f"Uploading {self.model_name} in {hf_hub_path}",
                repo_id=hf_hub_path,
                operations=hf_operations,
            )
            logger.info(
                f"Successfully uploaded {self.model_name} to {hf_hub_path} repo in HF hub!"
            )

        except:
            from huggingface_hub import create_repo

            repo_name = os.path.basename(os.path.normpath(hf_hub_path))
            logger.info(
                f"Creating {repo_name} in the HF hub since it does not exist..."
            )
            create_repo(repo_id=repo_name)
            logger.info(f"Successfully created {repo_name} in the HF hub!")

            api.create_commit(
                commit_message=f"Uploading {self.model_name} in {hf_hub_path}",
                repo_id=hf_hub_path,
                operations=hf_operations,
            )

        shutil.rmtree(tempdir)

    @classmethod
    def load_from_hf_hub(cls, hf_hub_path: str, allow_pickle=False):  # pragma: no cover
        """Class method to be used to load a pretrained model from the Hugging Face hub

        Args:
            hf_hub_path (str): The path where the model should have been be saved on the
                hugginface hub.

        .. note::
            This function requires the folder to contain:

            - | a ``model_config.json`` and a ``model.pt`` if no custom architectures were provided

            **or**

            - | a ``model_config.json``, a ``model.pt`` and a ``encoder.pkl`` (resp.
                ``decoder.pkl``) if a custom encoder (resp. decoder) was provided
        """

        if not hf_hub_is_available():
            raise ModuleNotFoundError(
                "`huggingface_hub` package must be installed to load models from the HF hub. "
                "Run `python -m pip install huggingface_hub` and log in to your account with "
                "`huggingface-cli login`."
            )

        else:
            from huggingface_hub import hf_hub_download

        logger.info(f"Downloading {cls.__name__} files for rebuilding...")

        _ = hf_hub_download(repo_id=hf_hub_path, filename="environment.json")
        config_path = hf_hub_download(repo_id=hf_hub_path, filename="model_config.json")
        dir_path = os.path.dirname(config_path)

        _ = hf_hub_download(repo_id=hf_hub_path, filename="model.pt")

        model_config = cls._load_model_config_from_folder(dir_path)

        if (
            cls.__name__ + "Config" != model_config.name
            and cls.__name__ + "_Config" != model_config.name
        ):
            warnings.warn(
                f"You are trying to load a "
                f"`{ cls.__name__}` while a "
                f"`{model_config.name}` is given."
            )

        model_weights = cls._load_model_weights_from_folder(dir_path)
        if (len(model_config.custom_architectures) >= 1) and not allow_pickle:
            warnings.warn(
                "You are about to download pickled files from the HF hub that may have "
                "been created by a third party and so could potentially harm your computer. If you "
                "are sure that you want to download them set `allow_pickle=true`."
            )

        else:
            custom_archi_dict = {}
            for archi in model_config.custom_architectures:
                _ = hf_hub_download(repo_id=hf_hub_path, filename=archi + ".pkl")
                archi_net = cls._load_custom_archi_from_folder(dir_path, archi)
                custom_archi_dict[archi] = archi_net
                logger.info(f"Successfully downloaded {archi} architecture.")

            logger.info(f"Successfully downloaded {cls.__name__} model!")

            model = cls(model_config, **custom_archi_dict)
            model.load_state_dict(model_weights)

            return model

    def generate_from_prior(self, n_samples, **kwargs):
        """
        Generate latent samples from the prior distribution.
        This is the base class in which we consider a static standard Normal Prior.
        This may be overwritten in subclasses.
        """
        sample_shape = (
            [n_samples, self.latent_dim] if n_samples > 1 else [self.latent_dim]
        )
        z = dist.Normal(0, 1).rsample(sample_shape).to(self.device)
        return ModelOutput(z=z, one_latent_space=True)

    def cond_nll_from_subset(
        self,
        inputs: MultimodalBaseDataset,
        subset: Union[list, tuple],
        pred_mods: Union[list, tuple],
        K=1000,
        batch_size_k=100,
    ):
        """
        Compute the negative log-likelihoods of the conditional generative models : encoding
        from one / a subset of modalities to generate others.
        log p(x_i|x_j,x_l) = log 1/K \sum_k p(x_i|z_k).


        Args:
            inputs (MultimodalBaseDataset): The inputs to use for the estimation.
            subset (list, tuple): The modalities to take as inputs.
            pred_mods (list, tuple): The modalities to take into account in the prediction.
            K (int, optional): The number of samples. Defaults to 1000.
            batch_size_k (int, optional): Batch size to use. Defaults to 100.
        """

        cnll = {m: [] for m in pred_mods}

        # Encode using the subset modalities
        nb_computed_samples = 0

        while nb_computed_samples < K:
            n_samples = min(batch_size_k, K - nb_computed_samples)
            encode_output = self.encode(inputs, subset, N=n_samples)
            if encode_output.one_latent_space:
                decode_output = self.decode(encode_output, pred_mods)

                for mod in pred_mods:
                    recon = decode_output[mod]  # (batch_size_k, n_data, *recon_size )
                    target = inputs.data[mod]
                    lpxz = self.recon_log_probs[mod](recon, target)
                    cnll[mod].append(torch.logsumexp(lpxz, dim=0))

            nb_computed_samples += n_samples

        cnll = {m: torch.logsumexp(torch.stack(cnll[mod]), dim=0) for m in cnll}
        return cnll



"""Training Callbacks for training monitoring integrated in `pythae` (inspired from 
https://github.com/huggingface/transformers/blob/master/src/transformers/trainer_callback.py)"""

import importlib
import json
import logging
import os
import warnings
from typing import Literal

import numpy as np
from tqdm.auto import tqdm

from ...models import BaseMultiVAEConfig
from .base_trainer_config import BaseTrainerConfig

logger = logging.getLogger(__name__)


def wandb_is_available():
    return importlib.util.find_spec("wandb") is not None


def mlflow_is_available():
    return importlib.util.find_spec("mlflow") is not None


def comet_is_available():
    return importlib.util.find_spec("comet_ml") is not None


def rename_logs(logs):
    train_prefix = "train_"
    eval_prefix = "eval_"

    clean_logs = {}

    for metric_name in logs.keys():
        if metric_name.startswith(train_prefix):
            clean_logs[metric_name.replace(train_prefix, "train/")] = logs[metric_name]

        if metric_name.startswith(eval_prefix):
            clean_logs[metric_name.replace(eval_prefix, "eval/")] = logs[metric_name]

    return clean_logs


class TrainingCallback:
    """
    Base class for creating training callbacks
    """

    def on_init_end(self, training_config: BaseTrainerConfig, **kwargs):
        """
        Event called at the end of the initialization of the [`Trainer`].
        """

    def on_train_begin(self, training_config: BaseTrainerConfig, **kwargs):
        """
        Event called at the beginning of training.
        """

    def on_train_end(self, training_config: BaseTrainerConfig, **kwargs):
        """
        Event called at the end of training.
        """

    def on_epoch_begin(self, training_config: BaseTrainerConfig, **kwargs):
        """
        Event called at the beginning of an epoch.
        """

    def on_epoch_end(self, training_config: BaseTrainerConfig, **kwargs):
        """
        Event called at the end of an epoch.
        """

    def on_train_step_begin(self, training_config: BaseTrainerConfig, **kwargs):
        """
        Event called at the beginning of a training step.
        """

    def on_train_step_end(self, training_config: BaseTrainerConfig, **kwargs):
        """
        Event called at the end of a training step.
        """

    def on_eval_step_begin(self, training_config: BaseTrainerConfig, **kwargs):
        """
        Event called at the beginning of a evaluation step.
        """

    def on_eval_step_end(self, training_config: BaseTrainerConfig, **kwargs):
        """
        Event called at the end of a evaluation step.
        """

    def on_evaluate(self, training_config: BaseTrainerConfig, **kwargs):
        """
        Event called after an evaluation phase.
        """

    def on_prediction_step(self, training_config: BaseTrainerConfig, **kwargs):
        """
        Event called after a prediction phase.
        """

    def on_save(self, training_config: BaseTrainerConfig, **kwargs):
        """
        Event called after a checkpoint save.
        """

    def on_save_checkpoint(self, training_config: BaseTrainerConfig, **kwargs):
        """
        Event called after a checkpoint save.
        """

    def on_log(self, training_config: BaseTrainerConfig, logs, **kwargs):
        """
        Event called after logging the last logs.
        """


class CallbackHandler:
    """
    Class to handle list of Callback.
    """

    def __init__(self, callbacks, model):
        self.callbacks = []
        for cb in callbacks:
            self.add_callback(cb)
        self.model = model

    def add_callback(self, callback):
        cb = callback() if isinstance(callback, type) else callback
        cb_class = callback if isinstance(callback, type) else callback.__class__
        if cb_class in [c.__class__ for c in self.callbacks]:
            logger.warning(
                f"You are adding a {cb_class} to the callbacks but one is already used."
                f" The current list of callbacks is\n: {self.callback_list}"
            )
        self.callbacks.append(cb)

    @property
    def callback_list(self):
        return "\n".join(cb.__class__.__name__ for cb in self.callbacks)

    @property
    def callback_list(self):
        return "\n".join(cb.__class__.__name__ for cb in self.callbacks)

    def on_init_end(self, training_config: BaseTrainerConfig, **kwargs):
        self.call_event("on_init_end", training_config, **kwargs)

    def on_train_step_begin(self, training_config: BaseTrainerConfig, **kwargs):
        self.call_event("on_train_step_begin", training_config, **kwargs)

    def on_train_step_end(self, training_config: BaseTrainerConfig, **kwargs):
        self.call_event("on_train_step_end", training_config, **kwargs)

    def on_eval_step_begin(self, training_config: BaseTrainerConfig, **kwargs):
        self.call_event("on_eval_step_begin", training_config, **kwargs)

    def on_eval_step_end(self, training_config: BaseTrainerConfig, **kwargs):
        self.call_event("on_eval_step_end", training_config, **kwargs)

    def on_train_begin(self, training_config: BaseTrainerConfig, **kwargs):
        self.call_event("on_train_begin", training_config, **kwargs)

    def on_train_end(self, training_config: BaseTrainerConfig, **kwargs):
        self.call_event("on_train_end", training_config, **kwargs)

    def on_epoch_begin(self, training_config: BaseTrainerConfig, **kwargs):
        self.call_event("on_epoch_begin", training_config, **kwargs)

    def on_epoch_end(self, training_config: BaseTrainerConfig, **kwargs):
        self.call_event("on_epoch_end", training_config)

    def on_evaluate(self, training_config: BaseTrainerConfig, **kwargs):
        self.call_event("on_evaluate", **kwargs)

    def on_save(self, training_config: BaseTrainerConfig, **kwargs):
        self.call_event("on_save", training_config, **kwargs)

    def on_save_checkpoint(self, training_config: BaseTrainerConfig, **kwargs):
        self.call_event("on_save_checkpoint", training_config, **kwargs)

    def on_log(self, training_config: BaseTrainerConfig, logs, **kwargs):
        self.call_event("on_log", training_config, logs=logs, **kwargs)

    def on_prediction_step(self, training_config: BaseTrainerConfig, **kwargs):
        self.call_event("on_prediction_step", training_config, **kwargs)

    def call_event(self, event, training_config, **kwargs):
        for callback in self.callbacks:
            result = getattr(callback, event)(
                training_config,
                model=self.model,
                **kwargs,
            )


class MetricConsolePrinterCallback(TrainingCallback):
    """
    A :class:`TrainingCallback` printing the training logs in the console.
    """

    def __init__(self):
        self.logger = logging.getLogger(__name__)

        # make it print to the console.
        console = logging.StreamHandler()
        self.logger.addHandler(console)
        self.logger.setLevel(logging.INFO)

    def on_log(self, training_config: BaseTrainerConfig, logs, **kwargs):
        logger = kwargs.pop("logger", self.logger)
        rank = kwargs.pop("rank", -1)

        if logger is not None and (rank == -1 or rank == 0):
            epoch_train_loss = logs.get("train_epoch_loss", None)
            epoch_eval_loss = logs.get("eval_epoch_loss", None)

            logger.info(
                "--------------------------------------------------------------------------"
            )
            if epoch_train_loss is not None:
                logger.info(f"Train loss: {np.round(epoch_train_loss, 4)}")
            if epoch_eval_loss is not None:
                logger.info(f"Eval loss: {np.round(epoch_eval_loss, 4)}")
            logger.info(
                "--------------------------------------------------------------------------"
            )


class ProgressBarCallback(TrainingCallback):
    """
    A :class:`TrainingCallback` printing the training progress bar.
    """

    def __init__(self):
        self.train_progress_bar = None
        self.eval_progress_bar = None

    def on_train_step_begin(self, training_config: BaseTrainerConfig, **kwargs):
        epoch = kwargs.pop("epoch", None)
        train_loader = kwargs.pop("train_loader", None)
        rank = kwargs.pop("rank", -1)
        if train_loader is not None:
            if rank == 0 or rank == -1:
                self.train_progress_bar = tqdm(
                    total=len(train_loader),
                    unit="batch",
                    desc=f"Training of epoch {epoch}/{training_config.num_epochs}",
                )

    def on_eval_step_begin(self, training_config: BaseTrainerConfig, **kwargs):
        epoch = kwargs.pop("epoch", None)
        eval_loader = kwargs.pop("eval_loader", None)
        rank = kwargs.pop("rank", -1)
        if eval_loader is not None:
            if rank == 0 or rank == -1:
                self.eval_progress_bar = tqdm(
                    total=len(eval_loader),
                    unit="batch",
                    desc=f"Eval of epoch {epoch}/{training_config.num_epochs}",
                )

    def on_train_step_end(self, training_config: BaseTrainerConfig, **kwargs):
        if self.train_progress_bar is not None:
            self.train_progress_bar.update(1)

    def on_eval_step_end(self, training_config: BaseTrainerConfig, **kwargs):
        if self.eval_progress_bar is not None:
            self.eval_progress_bar.update(1)

    def on_epoch_end(self, training_config: BaseTrainerConfig, **kwags):
        if self.train_progress_bar is not None:
            self.train_progress_bar.close()

        if self.eval_progress_bar is not None:
            self.eval_progress_bar.close()


class WandbCallback(TrainingCallback):  # pragma: no cover
    """
    A :class:`TrainingCallback` integrating the experiment tracking tool
    `wandb` (https://wandb.ai/).

    It allows users to store their configs, monitor their trainings
    and compare runs through a graphic interface. To be able use this feature you will need:

        - a valid `wandb` account
        - the package `wandb` installed in your virtual env. If not you can install it with

        .. code-block::

            $ pip install wandb

        - to be logged in to your wandb account using

        .. code-block::

            $ wandb login
    """

    def __init__(self):
        if not wandb_is_available():
            raise ModuleNotFoundError(
                "`wandb` package must be installed. Run `pip install wandb`"
            )

        else:
            import wandb

            self._wandb = wandb

    def setup(
        self,
        training_config: BaseTrainerConfig,
        model_config: BaseMultiVAEConfig = None,
        project_name: str = "multivae_experiment",
        entity_name: str = None,
        run_id: str = None,
        resume: Literal[
            "allow",
            "must",
        ] = "allow",
        **kwargs,
    ):
        """
        Setup the WandbCallback.

        args:
            training_config (BaseTrainerConfig): The training configuration used in the run.

            model_config (BaseMultiVAEConfig): The model configuration used in the run.

            project_name (str): The name of the wandb project to use.

            entity_name (str): The name of the wandb entity to use.

            run_id (str): If resume training, the id of the existing wandb_run

            resume (Literal) : wether to log on the provided run_id. Default to 'allow'.
        """

        self.is_initialized = True

        training_config_dict = training_config.to_dict()

        self.run = self._wandb.init(
            project=project_name, entity=entity_name, id=run_id, resume=resume
        )

        if model_config is not None:
            model_config_dict = model_config.to_dict()

            self._wandb.config.update(
                {
                    "training_config": training_config_dict,
                    "model_config": model_config_dict,
                }
            )

        else:
            self._wandb.config.update({**training_config_dict})

        self._wandb.define_metric("train/global_step")
        self._wandb.define_metric("*", step_metric="train/global_step", step_sync=True)

    def on_train_begin(self, training_config: BaseTrainerConfig, **kwargs):
        model_config = kwargs.pop("model_config", None)
        if not self.is_initialized:
            self.setup(training_config, model_config=model_config)

    def on_log(self, training_config: BaseTrainerConfig, logs, **kwargs):
        global_step = kwargs.pop("global_step", None)
        logs = rename_logs(logs)

        self._wandb.log({**logs, "train/global_step": global_step})

    def on_prediction_step(self, training_config: BaseTrainerConfig, **kwargs):
        kwargs.pop("global_step", None)

        column_names = ["images_id", "truth", "reconstruction", "normal_generation"]

        reconstructions = kwargs.pop("reconstructions", None)

        for cond_mod in reconstructions:
            image = self._wandb.Image(reconstructions[cond_mod])
            self._wandb.log({"recon_from_" + cond_mod: image})

    def on_save_checkpoint(self, training_config: BaseTrainerConfig, **kwargs):
        checkpoint_dir = kwargs.pop("checkpoint_dir", None)
        if checkpoint_dir is None:
            raise AttributeError(
                "wandb callback on_save_checkpoint is called without"
                "a checkpoint directory information. Please provide checkpoint_dir=.."
            )
        with open(os.path.join(checkpoint_dir, "checkpoint_info.json"), "r") as fp:
            info_dict = json.load(fp)
            info_dict["wandb_run"] = self._wandb.run.id
        with open(os.path.join(checkpoint_dir, "checkpoint_info.json"), "w") as fp:
            json.dump(info_dict, fp)

    def on_save(self, training_config: BaseTrainerConfig, **kwargs):
        dir_path = kwargs.pop("dir_path", None)
        if dir_path is None:
            warnings.warn(
                "wandb callback on_save is called without"
                "a  directory information. Please provide dir_path=.."
            )
            return
        info_dict = dict(
            entity_name=self.run.entity,
            project_name=self.run.project,
            id=self.run.id,
            path=self.run.path,
        )
        with open(os.path.join(dir_path, "wandb_info.json"), "w") as fp:
            json.dump(info_dict, fp)

    def on_train_end(self, training_config: BaseTrainerConfig, **kwargs):
        self.run.finish()



import pytest
from torch.utils.data import DataLoader

from multivae.data.datasets.celeba import CelebAttr, DatasetOutput


class Test:
    def test(self):
        try:
            dataset = CelebAttr("../data", "valid")
        except:
            return
        item = dataset[0]
        assert isinstance(item, DatasetOutput)
        assert hasattr(item, "data")
        assert hasattr(item, "labels")

        assert item.data["image"].shape == (3, 64, 64)
        assert item.data["attributes"].shape == (18,)

        dl = DataLoader(dataset, 12)
        batch = next(iter(dl))

        assert isinstance(batch, DatasetOutput)
        assert hasattr(batch, "data")
        assert hasattr(batch, "labels")

        assert batch.data["image"].shape == (12, 3, 64, 64)
        assert batch.data["attributes"].shape == (12, 18)



from pydantic.dataclasses import dataclass

from ..base.base_config import BaseMultiVAEConfig


@dataclass
class BaseJointModelConfig(BaseMultiVAEConfig):
    """
    This is the base config for joint models.



    """

    pass



"""
Implementation of the Variational Mixture-of-Experts Autoencoder model from the paper "Variational Mixture-of-Experts Autoencoders for
Multi-Modal Deep Generative Models"
(https://arxiv.org/abs/1911.03393)

"""

from .mmvaePlus_config import MMVAEPlusConfig
from .mmvaePlus_model import MMVAEPlus

__all__ = ["MMVAEPlus", "MMVAEPlusConfig"]



"""Base sampler class, adapted from Pythae Base Sampler."""


import logging
import os

import numpy as np
import torch
from imageio import imwrite

from ...models import BaseMultiVAE
from .base_sampler_config import BaseSamplerConfig

logger = logging.getLogger(__name__)

# make it print to the console.
console = logging.StreamHandler()
logger.addHandler(console)
logger.setLevel(logging.INFO)


class BaseSampler:
    """
    Base class for samplers used to generate from the MultiVae models' joint latent spaces.

    Args:
        model (BaseMultivae): The model to sample from.
        sampler_config (BaseSamplerConfig): An instance of BaseSamplerConfig in which any sampler's
            parameters is made available. If None a default configuration is used. Default: None
    """

    def __init__(self, model: BaseMultiVAE, sampler_config: BaseSamplerConfig = None):
        if sampler_config is None:
            sampler_config = BaseSamplerConfig()

        self.model = model
        self.model.eval()
        self.sampler_config = sampler_config
        self.is_fitted = False

        device = "cuda" if torch.cuda.is_available() else "cpu"

        self.device = device
        self.model.device = device

        self.model.to(device)
        self.name = "BaseSampler"

    def fit(self, *args, **kwargs):
        """Function to be called to fit the sampler before sampling"""
        pass

    def sample(
        self,
        n_samples: int = 1,
        batch_size: int = 500,
        return_gen: bool = True,
    ):
        """Main sampling function of the sampler.

        Args:
            num_samples (int): The number of samples to generate
            batch_size (int): The batch size to use during sampling
            return_gen (bool): Whether the sampler should directly return a the generated
                data. Default: True.

        Returns:
            ~torch.Tensor: The generated images
        """
        raise NotImplementedError()

    def save(self, dir_path):
        """Method to save the sampler config. The config is saved a as ``sampler_config.json``
        file in ``dir_path``"""

        self.sampler_config.save_json(dir_path, "sampler_config")



from typing import Any, Dict

import torch
from pythae.data.datasets import DatasetOutput


def set_inputs_to_device(inputs: Dict[str, Any], device: str = "cpu"):
    inputs_on_device = inputs

    if device == "cuda":
        cuda_inputs = dict.fromkeys(inputs)

        for key in inputs.keys():
            if torch.is_tensor(inputs[key]):
                cuda_inputs[key] = inputs[key].cuda()

            elif isinstance(inputs[key], dict):
                cuda_inputs[key] = dict.fromkeys(inputs[key])
                for subkey in inputs[key].keys():
                    if torch.is_tensor(inputs[key][subkey]):
                        cuda_inputs[key][subkey] = inputs[key][subkey].cuda()
                    elif isinstance(inputs[key][subkey], dict):
                        cuda_inputs[key][subkey] = dict.fromkeys(inputs[key][subkey])
                        for subsubkey in inputs[key][subkey].keys():
                            if torch.is_tensor(inputs[key][subkey][subsubkey]):
                                cuda_inputs[key][subkey][subsubkey] = inputs[key][subkey][subsubkey].cuda()
                            else:
                                cuda_inputs[key][subkey][subsubkey] = inputs[key][subkey][subsubkey]
                    else:
                        cuda_inputs[key][subkey] = inputs[key][subkey]
            else:
                cuda_inputs[key] = inputs[key]
        inputs_on_device = cuda_inputs

    return DatasetOutput(**inputs_on_device)



import os

import numpy as np
import pytest
import torch
from pythae.models.base import BaseAEConfig
from pythae.models.nn.benchmarks.mnist.convnets import (
    Decoder_Conv_AE_MNIST,
    Encoder_Conv_AE_MNIST,
)
from pythae.models.nn.default_architectures import Encoder_AE_MLP, Encoder_VAE_MLP
from torch import nn

from multivae.models.base import BaseMultiVAE, BaseMultiVAEConfig
from multivae.models.joint_models import BaseJointModel
from multivae.models.nn.default_architectures import (
    Decoder_AE_MLP,
    MultipleHeadJointEncoder,
)


class Test_JointModel:
    @pytest.fixture
    def input_model(self):
        model_config = BaseMultiVAEConfig(n_modalities=2, latent_dim=10)
        config1 = BaseAEConfig(input_dim=(7,), latent_dim=10)
        config2 = BaseAEConfig(input_dim=(3,), latent_dim=10)

        encoders = dict(mod1=Encoder_AE_MLP(config1), mod2=Encoder_AE_MLP(config2))

        decoders = dict(mod1=Decoder_AE_MLP(config1), mod2=Decoder_AE_MLP(config2))

        data = dict(mod1=torch.ones((3, 7)), mod2=torch.ones((3, 3)))
        return dict(encoders=encoders, model_config=model_config, decoders=decoders)

    def test(self, input_model):
        model = BaseJointModel(**input_model)

        assert type(model.encoders) == nn.ModuleDict
        assert isinstance(model.encoders["mod1"], Encoder_AE_MLP)
        assert isinstance(model.encoders["mod2"], Encoder_AE_MLP)
        assert type(model.decoders) == nn.ModuleDict
        assert isinstance(model.decoders["mod1"], Decoder_AE_MLP)
        assert isinstance(model.decoders["mod2"], Decoder_AE_MLP)

        assert isinstance(model.joint_encoder, MultipleHeadJointEncoder)
        assert model.latent_dim == input_model["model_config"].latent_dim

    @pytest.fixture
    def input2(self):
        model_config = BaseMultiVAEConfig(
            n_modalities=2, latent_dim=10, input_dims=dict(mod1=(7,), mod2=(3,))
        )

        return dict(model_config=model_config)

    def test2(self, input2):
        model = BaseJointModel(**input2)

        assert type(model.encoders) == nn.ModuleDict
        assert isinstance(model.encoders["mod1"], Encoder_VAE_MLP)
        assert isinstance(model.encoders["mod2"], Encoder_VAE_MLP)
        assert model.encoders["mod1"].input_dim == (7,)
        assert model.encoders["mod2"].input_dim == (3,)
        assert model.encoders["mod1"].latent_dim == 10
        assert model.encoders["mod2"].latent_dim == 10

        assert type(model.decoders) == nn.ModuleDict
        assert isinstance(model.decoders["mod1"], Decoder_AE_MLP)
        assert isinstance(model.decoders["mod2"], Decoder_AE_MLP)

        assert isinstance(model.joint_encoder, MultipleHeadJointEncoder)
        assert model.latent_dim == input2["model_config"].latent_dim



import random

import numpy as np
import torch


def set_seed(seed: int):
    """
    Functions setting the seed for reproducibility on ``random``, ``numpy``,
    and ``torch``

    Args:

        seed (int): The seed to be applied
    """

    random.seed(seed)
    np.random.seed(seed)
    torch.manual_seed(seed)
    torch.cuda.manual_seed_all(seed)


def update_dict(dict1, dict2):
    """Modify in place the first dict by adding values of the second dict"""
    for k in dict2.keys():
        if k in dict1.keys():
            dict1[k] += dict2[k]
        else:
            dict1[k] = dict2[k]



import logging

import torch
from sklearn import mixture
from torch.utils.data import DataLoader

from multivae.data import MultimodalBaseDataset
from multivae.data.utils import set_inputs_to_device
from multivae.models.base import ModelOutput

from ...models import BaseMultiVAE
from ..base import BaseSampler
from .gaussian_mixture_config import GaussianMixtureSamplerConfig

logger = logging.getLogger(__name__)

# make it print to the console.
console = logging.StreamHandler()
logger.addHandler(console)
logger.setLevel(logging.INFO)


class GaussianMixtureSampler(BaseSampler):
    """
    
    Fits a Gaussian Mixture in the Multimodal Autoencoder's latent space.
    If the model has several latent spaces, it fits a gaussian mixture per latent space.

    Args:
        model (BaseMultiVAE): The model to sample from.
        sampler_config (BaseSamplerConfig): An instance of BaseSamplerConfig in which any sampler's
            parameters is made available. If None a default configuration is used. Default: None.

    .. note::

        The method :class:`~multivae.samplers.GaussianMixtureSampler.fit` must be called to fit the sampler
        before sampling.

    """

    def __init__(
        self, model: BaseMultiVAE, sampler_config: GaussianMixtureSamplerConfig = None
    ):
        if sampler_config is None:
            sampler_config = GaussianMixtureSamplerConfig()

        BaseSampler.__init__(self, model=model, sampler_config=sampler_config)
        self.n_components = sampler_config.n_components
        self.device = "cuda" if torch.cuda.is_available() else "cpu"
        self.model.to(self.device)
        self.name = "GaussianMixtureSampler"

    def fit(self, train_data: MultimodalBaseDataset, **kwargs):
        """Method to fit the sampler from the training data

        Args:
            train_data (MultimodalBaseDataset): The train data needed to retreive the training embeddings
                    and fit the mixture in the latent space. Must be an instance of MultimodalBaseDataset.
        """

        train_loader = DataLoader(
            dataset=train_data,
            batch_size=100,
            shuffle=False,
        )

        z = []
        if self.model.multiple_latent_spaces:
            mod_z = {m: [] for m in self.model.encoders}

        # Compute all embeddings
        with torch.no_grad():
            for _, inputs in enumerate(train_loader):
                inputs = set_inputs_to_device(inputs, self.device)
                output = self.model.encode(inputs)
                z.append(output.z)

                if self.model.multiple_latent_spaces:
                    for m in mod_z:
                        mod_z[m].append(output.modalities_z[m])

        z = torch.cat(z)
        if self.model.multiple_latent_spaces:
            mod_z = {m: torch.cat(mod_z[m]) for m in mod_z}

        if self.n_components > z.shape[0]:
            self.n_components = z.shape[0]
            logger.warning(
                f"Setting the number of component to {z.shape[0]} since"
                "n_components > n_samples when fitting the gmm"
            )

        gmm = mixture.GaussianMixture(
            n_components=self.n_components,
            covariance_type="full",
            max_iter=2000,
            verbose=0,
            tol=1e-3,
        )
        gmm.fit(z.cpu().detach())
        self.gmm = gmm

        if self.model.multiple_latent_spaces:
            self.mod_gmms = dict()
            for m in mod_z:
                gmm = mixture.GaussianMixture(
                    n_components=self.n_components,
                    covariance_type="full",
                    max_iter=2000,
                    verbose=0,
                    tol=1e-3,
                )

                gmm.fit(mod_z[m].cpu().detach())
                self.mod_gmms[m] = gmm

        self.is_fitted = True

    def sample(
        self, n_samples: int = 1, batch_size: int = 500, **kwargs
    ) -> torch.Tensor:
        """Main sampling function of the sampler.

        Args:
            num_samples (int): The number of samples to generate
            batch_size (int): The batch size to use during sampling
            save_sampler_config (bool): Whether to save the sampler config. It is saved in
                output_dir

        Returns:
            ModelOutput similar as the one returned by the encode function or generate_from_prior function.
        """

        if not self.is_fitted:
            raise ArithmeticError(
                "The sampler needs to be fitted by calling sampler.fit() method"
                "before sampling."
            )

        full_batch_nbr = int(n_samples / batch_size)
        last_batch_samples_nbr = n_samples % batch_size

        batches_sizes = [batch_size] * full_batch_nbr
        if last_batch_samples_nbr != 0:
            batches_sizes = batches_sizes + [last_batch_samples_nbr]

        z_list = []

        if self.model.multiple_latent_spaces:
            mod_z = {m: [] for m in self.model.encoders}

        for batch_size in batches_sizes:
            z = (
                torch.tensor(self.gmm.sample(batch_size)[0])
                .to(self.device)
                .type(torch.float)
            )

            z_list.append(z)

            if self.model.multiple_latent_spaces:
                for m in mod_z:
                    z = (
                        torch.tensor(self.mod_gmms[m].sample(batch_size)[0])
                        .to(self.device)
                        .type(torch.float)
                    )
                    mod_z[m].append(z)

        output = ModelOutput(z=torch.cat(z_list, dim=0))

        if self.model.multiple_latent_spaces:
            output["one_latent_space"] = False
            output["modalities_z"] = {m: torch.cat(mod_z[m]) for m in mod_z}
        else:
            output["one_latent_space"] = True

        return output



import logging
import os
import pickle
from collections import defaultdict
from pathlib import Path
from typing import Tuple, Union

import numpy as np
import pandas as pd
import torch
from nltk.tokenize import RegexpTokenizer
from PIL import Image
from pythae.data.datasets import DatasetOutput
from torchvision import transforms
from torchvision.datasets import MNIST, SVHN

from .base import MultimodalBaseDataset
from .utils import ResampleDataset

logger = logging.getLogger(__name__)
console = logging.StreamHandler()
logger.addHandler(console)
logger.setLevel(logging.INFO)

class CUB(MultimodalBaseDataset): # pragma: no cover
    """

    A paired text img CUB dataset.

    Args:
        path (str) : The path where the data is saved.
        split (str) : Either 'train' or 'test'. Default: 'train'
        captions_per_image (int): The number of captions text per image. Default: 10
        max_words_in_caption (int): The number of words in the captions. Default: 18
        im_size (Tuple[int]): The desired size of the images. Default: (64, 64)
        img_transform (Transforms): The transformations to be applied to the images. If 
            None, nothing is done. Default: None.
    """

    def __init__(
        self,
        data_path: Union[str, Path],
        split: str = "train",
        captions_per_image: int = 10,
        max_words_in_caption: int = 18,
        im_size: Tuple[int] = (64, 64),
        img_transform=None,
    ):  
        if split not in ["train", "test"]:
            raise AttributeError("Possible values for split are 'train' or 'test'")
        
        if captions_per_image > 10:
            raise AttributeError("Maximum number of captions per image is 10.")
        
        self.img_transform = img_transform
        self.normalize = transforms.Compose([
            transforms.ToTensor(),
            transforms.Normalize((0.5, 0.5, 0.5), (0.5, 0.5, 0.5))])
        self.captions_per_img = captions_per_image
        self.data_path = data_path
        self.split = split
        self.imsize = im_size
        self.max_words_in_captions = max_words_in_caption
        self.tokenizer = RegexpTokenizer(r'\w+')

        self.train_test_split = self._load_train_test_split()
        filenames = self._load_filenames()
        labels = self._load_labels()

        # get train_test_filenames
        train_filenames = filenames[self.train_test_split.split==1]
        test_filenames = filenames[self.train_test_split.split==0]
        train_filenames.reset_index(drop=True, inplace=True)
        test_filenames.reset_index(drop=True, inplace=True)

        # get train_test_labels
        train_labels = labels[self.train_test_split.split==1]
        test_labels = labels[self.train_test_split.split==0]
        train_labels.reset_index(drop=True, inplace=True)
        test_labels.reset_index(drop=True, inplace=True)

        train_captions = self._load_captions(train_filenames.name)
        test_captions = self._load_captions(test_filenames.name)
        self.bbox = self._load_bbox() 

        train_captions_new, test_captions_new, idxtoword, wordtoidx, vocab_size = self.build_vocab(
            train_captions,
            test_captions
        )
        
        self.train_filenames = train_filenames
        self.test_filenames = test_filenames
        self.train_labels = train_labels
        self.test_labels = test_labels
        self.train_captions = train_captions_new
        self.test_captions = test_captions_new
        self.idxtoword = idxtoword
        self.wordtoidx = wordtoidx
        self.vocab_size = vocab_size
        self.labels = None

    def _load_train_test_split(self):
        train_test_split = pd.read_csv(
            os.path.join(self.data_path, "train_test_split.txt"),
            delim_whitespace=True,
            header=None,
            names=["id", "split"]
        )

        return train_test_split
    
    def _load_filenames(self):
        filenames = pd.read_csv(
            os.path.join(self.data_path, "images.txt"),
            delim_whitespace=True,
            header=None,
            names=["id", "name"]
        )
        filenames["name"] = filenames["name"].str.replace('.jpg', '')
    
        return filenames


    def _load_captions(self, filenames):
        all_captions = defaultdict(list)
        for filename in filenames:
            cap_path = os.path.join(f'{self.data_path}', 'text', f'{filename}.txt')
            with open(cap_path, "r") as f:
                captions = f.read().splitlines()

                cnt = 0
                for cap in captions:
                    if len(cap) == 0:
                        continue
                    cap = cap.replace("\ufffd\ufffd", " ")
                    # picks out sequences of alphanumeric characters as tokens
                    # and drops everything else
                    tokenizer = RegexpTokenizer(r'\w+')
                    tokens = tokenizer.tokenize(cap.lower())
                    if len(tokens) == 0:
                        continue

                    tokens_new = []
                    for t in tokens:
                        t = t.encode('ascii', 'ignore').decode('ascii')
                        if len(t) > 0:
                            tokens_new.append(t)
                    all_captions[filename].append(tokens_new)
                    cnt += 1
                    if cnt == self.captions_per_img:
                        break
                if cnt < self.captions_per_img:
                    logger.error('ERROR: the captions for %s less than %d'
                          % (filename, cnt))

        return all_captions

    def _load_bbox(self):
        data_dir = self.data_path
        bbox_path = os.path.join(data_dir, 'bounding_boxes.txt')
        df_bounding_boxes = pd.read_csv(bbox_path,
                                        delim_whitespace=True,
                                        header=None).astype(int)

        filepath = os.path.join(data_dir, 'images.txt')
        df_filenames = \
            pd.read_csv(filepath, delim_whitespace=True, header=None)
        filenames = df_filenames[1].tolist()

        filename_bbox = {img_file[:-4]: [] for img_file in filenames}
        numImgs = len(filenames)
        for i in range(0, numImgs):
            bbox = df_bounding_boxes.iloc[i][1:].tolist()

            key = filenames[i][:-4]
            filename_bbox[key] = bbox

        return filename_bbox


    def get_imgs(self, img_path, bbox=None):
        img = Image.open(img_path).convert('RGB')
        width, height = img.size
        if bbox is not None:
            r = int(np.maximum(bbox[2], bbox[3]) * 0.75)
            center_x = int((2 * bbox[0] + bbox[2]) / 2)
            center_y = int((2 * bbox[1] + bbox[3]) / 2)
            y1 = np.maximum(0, center_y - r)
            y2 = np.minimum(height, center_y + r)
            x1 = np.maximum(0, center_x - r)
            x2 = np.minimum(width, center_x + r)
            img = img.crop([x1, y1, x2, y2])

        if self.img_transform is not None:
            img = self.img_transform(img)
            
        re_img = transforms.Resize(self.imsize)(img)
        ret = self.normalize(re_img)

        return ret
    
    def _load_labels(self):
        labels = pd.read_csv(
            os.path.join(self.data_path, "image_class_labels.txt"),
            delim_whitespace=True,
            header=None,
            names=["label"]
        )

        labels.reset_index(drop=True, inplace=True)

        return labels

    def build_vocab(self, train_captions, test_captions):
        word_counts = defaultdict(float)
        captions = [cap for cap_list in train_captions.values() for cap in cap_list] + \
            [cap for cap_list in test_captions.values() for cap in cap_list]
        for sent in captions:
            for word in sent:
                word_counts[word] += 1

        vocab = [w for w in word_counts if word_counts[w] >= 0]

        ixtoword = {}
        ixtoword[0] = '<end>'
        ixtoword[1] = '<pad>'
        wordtoix = {}
        wordtoix['<end>'] = 0
        wordtoix['<pad>'] = 1
        ix = 2
        for w in vocab:
            wordtoix[w] = ix
            ixtoword[ix] = w
            ix += 1

        train_captions_new = defaultdict(list)
        for cap_key in train_captions.keys():
            for t in train_captions[cap_key]:
                rev = []
                for w in t:
                    if w in wordtoix:
                        rev.append(wordtoix[w])
                # rev.append(0)  # do not need '<end>' token
                train_captions_new[cap_key].append(rev)

        test_captions_new = defaultdict(list)
        for cap_key in test_captions.keys():
            for t in test_captions[cap_key]:
                rev = []
                for w in t:
                    if w in wordtoix:
                        rev.append(wordtoix[w])
                # rev.append(0)  # do not need '<end>' token
                test_captions_new[cap_key].append(rev)

        return [train_captions_new, test_captions_new,
                ixtoword, wordtoix, len(ixtoword)]

    def get_caption(self, sent_ix, captions):
        # a list of indices for a sentence
        sent_caption = np.asarray(captions[sent_ix]).astype('int64')
        if (sent_caption == 0).sum() > 0:
            logger.error('ERROR: do not need END (0) token', sent_caption)
        num_words = len(sent_caption)
        # pad with 1s (i.e., '<pad>')
        x = np.ones((self.max_words_in_captions, 1), dtype='int64')
        x_len = num_words
        if num_words < self.max_words_in_captions:
            x[:num_words, 0] = sent_caption
            x[num_words, 0] = 0
        else:
            ix = list(np.arange(num_words))  # 1, 2, 3,..., maxNum
            np.random.shuffle(ix)
            ix = ix[:self.max_words_in_captions]
            ix = np.sort(ix)
            x[:, 0] = sent_caption[ix]
            x_len = self.max_words_in_captions
        padding_mask = x == 1
        return x, x_len, ~padding_mask
    
    def __len__(self):
        if self.split == "train":
            return len(self.train_filenames.name)
        
        return len(self.test_filenames.name)

    def __getitem__(self, index):
        if self.split == "train":
            names = self.train_filenames.name[index]
            captions = self.train_captions[names]
            labels = self.train_labels.label[index]

        else:
            names = self.test_filenames.name[index]
            captions = self.test_captions[names]
            labels = self.test_labels.label[index]
            
        
        bbox = self.bbox[names]
        img_path = os.path.join(self.data_path, 'images', f'{names}.jpg')
        imgs = self.get_imgs(img_path, bbox=bbox)

        sent_ix = np.random.randint(0, self.captions_per_img)

        #new_sent_ix = index * self.captions_per_img + sent_ix
        caps, cap_len, padding_mask = self.get_caption(sent_ix, captions)

        X = dict(img=imgs, text=dict(
            tokens=torch.tensor(caps).squeeze(-1),
            padding_mask=torch.FloatTensor(padding_mask).squeeze(-1),
            )
        )
        

        return DatasetOutput(data=X, labels=labels)


import json
import logging
import os

import torch.nn as nn

from ..base.base_utils import hf_hub_is_available

logger = logging.getLogger(__name__)
console = logging.StreamHandler()
logger.addHandler(console)
logger.setLevel(logging.INFO)


class AutoModel(nn.Module):
    "Utils class allowing to reload any :class:`multivae.models` automatically"

    def __init__(self) -> None:
        super().__init__()

    @classmethod
    def load_from_folder(cls, dir_path: str):
        """Class method to be used to load the model from a specific folder

        Args:
            dir_path (str): The path where the model should have been be saved.

        .. note::
            This function requires the folder to contain:

            - | a ``model_config.json`` and a ``model.pt`` if no custom architectures were provided

            **or**

            - | a ``model_config.json``, a ``model.pt`` and a ``encoders.pkl`` (resp.
                ``decoders.pkl``) if a custom encoders (resp. decoders) were provided
        """
        with open(os.path.join(dir_path, "model_config.json")) as f:
            model_name = json.load(f)["name"]

        if model_name == "JMVAEConfig":
            from ..jmvae import JMVAE

            model = JMVAE.load_from_folder(dir_path=dir_path)

        elif model_name == "JNFConfig":
            from ..jnf import JNF

            model = JNF.load_from_folder(dir_path=dir_path)

        elif model_name == "MMVAEConfig":
            from ..mmvae import MMVAE

            model = MMVAE.load_from_folder(dir_path=dir_path)
        elif model_name == "TELBOConfig":
            from ..telbo import TELBO

            model = TELBO.load_from_folder(dir_path)

        elif model_name == "MVAEConfig":
            from ..mvae import MVAE

            model = MVAE.load_from_folder(dir_path)

        elif model_name == "JNFDccaConfig":
            from ..jnf_dcca import JNFDcca

            model = JNFDcca.load_from_folder(dir_path)

        elif model_name == "MoPoEConfig":
            from ..mopoe import MoPoE

            model = MoPoE.load_from_folder(dir_path)

        elif model_name == "MVTCAEConfig":
            from ..mvtcae import MVTCAE

            model = MVTCAE.load_from_folder(dir_path)

        elif model_name == "MMVAEPlusConfig":
            from ..mmvaePlus import MMVAEPlus

            model = MMVAEPlus.load_from_folder(dir_path)

        else:
            raise NameError(
                "Cannot reload automatically the model... "
                f"The model name in the `model_config.json may be corrupted. Got {model_name}"
            )

        return model

    @classmethod
    def load_from_hf_hub(
        cls, hf_hub_path: str, allow_pickle: bool = False
    ):  # pragma: no cover
        """Class method to be used to load a automaticaly a pretrained model from the Hugging Face
        hub

        Args:
            hf_hub_path (str): The path where the model should have been be saved on the
                hugginface hub.

        .. note::
            This function requires the folder to contain:

            - | a ``model_config.json`` and a ``model.pt`` if no custom architectures were provided

            **or**

            - | a ``model_config.json``, a ``model.pt`` and a ``encoder.pkl`` (resp.
                ``decoder.pkl``) if a custom encoder (resp. decoder) was provided
        """

        if not hf_hub_is_available():
            raise ModuleNotFoundError(
                "`huggingface_hub` package must be installed to load models from the HF hub. "
                "Run `python -m pip install huggingface_hub` and log in to your account with "
                "`huggingface-cli login`."
            )

        else:
            from huggingface_hub import hf_hub_download

        logger.info(f"Downloading config file ...")

        config_path = hf_hub_download(repo_id=hf_hub_path, filename="model_config.json")
        dir_path = os.path.dirname(config_path)

        with open(os.path.join(dir_path, "model_config.json")) as f:
            model_name = json.load(f)["name"]

        if model_name == "JMVAEConfig":
            from ..jmvae import JMVAE

            model = JMVAE.load_from_hf_hub(hf_hub_path, allow_pickle)

        elif model_name == "JNFConfig":
            from ..jnf import JNF

            model = JNF.load_from_hf_hub(hf_hub_path, allow_pickle)

        elif model_name == "MMVAEConfig":
            from ..mmvae import MMVAE

            model = MMVAE.load_from_hf_hub(hf_hub_path, allow_pickle)
        elif model_name == "TELBOConfig":
            from ..telbo import TELBO

            model = TELBO.load_from_hf_hub(hf_hub_path, allow_pickle)

        elif model_name == "MVAEConfig":
            from ..mvae import MVAE

            model = MVAE.load_from_hf_hub(hf_hub_path, allow_pickle)

        elif model_name == "JNFDccaConfig":
            from ..jnf_dcca import JNFDcca

            model = JNFDcca.load_from_hf_hub(hf_hub_path, allow_pickle)

        elif model_name == "MoPoEConfig":
            from ..mopoe import MoPoE

            model = MoPoE.load_from_hf_hub(hf_hub_path, allow_pickle)

        elif model_name == "MVTCAEConfig":
            from ..mvtcae import MVTCAE

            model = MVTCAE.load_from_hf_hub(hf_hub_path, allow_pickle)

        elif model_name == "MMVAEPlusConfig":
            from ..mmvaePlus import MMVAEPlus

            model = MMVAEPlus.load_from_hf_hub(hf_hub_path, allow_pickle)

        else:
            raise NameError(
                "Cannot reload automatically the model... "
                f"The model name in the `model_config.json may be corrupted. Got {model_name}"
            )

        return model



from pydantic.dataclasses import dataclass

from ..base.evaluator_config import EvaluatorConfig


@dataclass
class LikelihoodsEvaluatorConfig(EvaluatorConfig):
    """

    Config class for the evaluation of the coherences module.

    Args:
        batch_size (int) : The batch size to use in the evaluation.  Default to 512
        wandb_path (str) : The user can provide the path of the wandb run with a
            format 'entity/projet_name/run_id' where the metrics should be logged.
            For an existing run (the training run), the info can be found in the training dir
                (in wandb_info.json)
            at the end of training (if wandb was used) or on the hugging_face webpage of the run.
            Otherwise the user can create a new wandb run and get the path with :

                .. code-block:: python

                    >>> import wandb
                    >>> run = wandb.init(entity = your_entity, project=your_project)
                    >>> wandb_path = run.path

            If None are provided, the metrics are not logged on wandb.
            Default to None.

        num_samples (int) : How many samples to use for likelihoods estimates. Default to 1000.
        batch_size_k (int) : How to batch the K samples for likelihoods estimates. Default to 100.
        unified_implementation (bool) : When the paper implementation of the likelihood differ from
            the unified implementation, specify which to use. Default to True.


    """

    num_samples: int = 1000
    batch_size_k: int = 100
    unified_implementation: bool = True



import os

import pytest

from multivae.models import BaseMultiVAEConfig


class Test_BaseMultiVAEConfig:
    @pytest.fixture(
        params=[
            dict(
                n_modalities=3,
                latent_dim=10,
                decoders_dist=dict(mod1="laplace", mod2="laplace"),
                decoder_dist_params=dict(mod1={"scale": 0.75}, mod2={"scale": 0.75}),
            ),
            dict(
                n_modalities=2,
                latent_dim=5,
                input_dims=dict(mod1=(2,), mod2=(3,)),
                decoders_dist=dict(mod1="laplace", mod2="laplace"),
                decoder_dist_params=dict(mod1={"scale": 0.75}, mod2={"scale": 0.75}),
            ),
        ]
    )
    def input_latent_pairs(self, request):
        return request.param

    def test_create_config(self, input_latent_pairs):
        config = BaseMultiVAEConfig(**input_latent_pairs)

        assert config.n_modalities == input_latent_pairs["n_modalities"]
        assert config.latent_dim == input_latent_pairs["latent_dim"]
        assert config.decoders_dist == input_latent_pairs["decoders_dist"]
        assert config.decoder_dist_params == input_latent_pairs["decoder_dist_params"]



import logging
from typing import Union

import numpy as np
import torch
import torch.distributions as dist
import torch.nn.functional as F
from pythae.models.base.base_utils import ModelOutput
from torch.distributions import Laplace, Normal

from multivae.data.datasets.base import MultimodalBaseDataset

from ..base import BaseMultiVAE
from .mmvae_config import MMVAEConfig

logger = logging.getLogger(__name__)
console = logging.StreamHandler()
logger.addHandler(console)
logger.setLevel(logging.INFO)


class MMVAE(BaseMultiVAE):

    """
    The Variational Mixture-of-Experts Autoencoder model.

    Args:
        model_config (MMVAEConfig): An instance of MMVAEConfig in which any model's
            parameters is made available.

        encoders (Dict[str, ~pythae.models.nn.base_architectures.BaseEncoder]): A dictionary containing
            the modalities names and the encoders for each modality. Each encoder is an instance of
            Pythae's BaseEncoder. Default: None.

        decoders (Dict[str, ~pythae.models.nn.base_architectures.BaseDecoder]): A dictionary containing
            the modalities names and the decoders for each modality. Each decoder is an instance of
            Pythae's BaseDecoder.
    """

    def __init__(
        self, model_config: MMVAEConfig, encoders: dict = None, decoders: dict = None
    ):
        super().__init__(model_config, encoders, decoders)

        self.K = model_config.K
        if model_config.prior_and_posterior_dist == "laplace_with_softmax":
            self.post_dist = Laplace
            self.prior_dist = Laplace
        elif model_config.prior_and_posterior_dist == "normal":
            self.post_dist = Normal
            self.prior_dist = Normal
        else:
            raise AttributeError(
                " The posterior_dist parameter must be "
                " either 'laplace_with_softmax' or 'normal'. "
                f" {model_config.prior_and_posterior_dist} was provided."
            )

        self.prior_mean = torch.nn.Parameter(
            torch.zeros(1, self.latent_dim), requires_grad=False
        )
        self.prior_log_var = torch.nn.Parameter(
            torch.zeros(1, self.latent_dim), requires_grad=model_config.learn_prior
        )

        self.model_name = "MMVAE"

    def log_var_to_std(self, log_var):
        """
        For latent distributions parameters, transform the log covariance to the
        standard deviation of the distribution either applying softmax or not.
        This follows the original implementation.
        """

        if self.model_config.prior_and_posterior_dist == "laplace_with_softmax":
            return F.softmax(log_var, dim=-1) * log_var.size(-1) + 1e-6
        else:
            return torch.exp(0.5 * log_var)

    @property
    def pz_params(self):
        """From the prior mean and log_covariance, return the mean and standard
        deviation, either applying softmax or not depending on the choice of prior
        distribution.

        Returns:
            tuple: mean, std
        """
        mean = self.prior_mean
        if self.model_config.prior_and_posterior_dist == "laplace_with_softmax":
            std = (
                F.softmax(self.prior_log_var, dim=-1) * self.prior_log_var.size(-1)
                + 1e-6
            )
        else:
            std = torch.exp(0.5 * self.prior_log_var)
        return mean, std

    def forward(self, inputs: MultimodalBaseDataset, **kwargs):
        # TODO : maybe implement a minibatch strategy for stashing the gradients before
        # backpropagation when using a large number k.
        # Also, I've only implemented the dreg_looser loss but it may be nice to offer other options.

        # First compute all the encodings for all modalities
        embeddings = {}
        qz_xs = {}
        qz_xs_detach = {}
        reconstructions = {}

        compute_loss = kwargs.pop("compute_loss", True)
        detailed_output = kwargs.pop("detailed_output", False)
        K = kwargs.pop("K", self.K)

        for cond_mod in self.encoders:
            output = self.encoders[cond_mod](inputs.data[cond_mod])
            mu, log_var = output.embedding, output.log_covariance

            sigma = self.log_var_to_std(log_var)

            qz_x = self.post_dist(mu, sigma)
            z_x = qz_x.rsample([K])

            # The DREG loss uses detached parameters in the loss computation afterwards.
            qz_x_detach = self.post_dist(mu.detach(), sigma.detach())

            # Then compute all the cross-modal reconstructions
            reconstructions[cond_mod] = {}
            for recon_mod in self.decoders:
                decoder = self.decoders[recon_mod]
                recon = decoder(z_x)["reconstruction"]

                reconstructions[cond_mod][recon_mod] = recon

            qz_xs[cond_mod] = qz_x
            embeddings[cond_mod] = z_x
            qz_xs_detach[cond_mod] = qz_x_detach

        # Compute DREG loss
        if compute_loss:
            # TODO : change
            loss_output = self.dreg_looser(
                qz_xs_detach, embeddings, reconstructions, inputs
            )

        else:
            loss_output = ModelOutput()
        if detailed_output:
            loss_output["qz_xs"] = qz_xs
            loss_output["qz_xs_detach"] = qz_xs_detach
            loss_output["zss"] = embeddings
            loss_output["recon"] = reconstructions

        return loss_output

    def dreg_looser(self, qz_xs, embeddings, reconstructions, inputs):
        if hasattr(inputs, "masks"):
            # Compute the number of available modalities per sample
            n_mods_sample = torch.sum(
                torch.stack(tuple(inputs.masks.values())).int(), dim=0
            )
        else:
            n_mods_sample = torch.tensor([self.n_modalities])

        lws = []
        zss = []
        for mod in embeddings:
            z = embeddings[mod]  # (K, n_batch, latent_dim)
            n_mods_sample = n_mods_sample.to(z.device)
            prior = self.prior_dist(*self.pz_params)
            lpz = prior.log_prob(z).sum(-1)

            if hasattr(inputs, "masks"):
                lqz_x = torch.stack(
                    [
                        qz_xs[m].log_prob(z).sum(-1) * inputs.masks[m].float()
                        for m in qz_xs
                    ]
                )  # n_modalities,K,nbatch
            else:
                lqz_x = torch.stack(
                    [qz_xs[m].log_prob(z).sum(-1) for m in qz_xs]
                )  # n_modalities,K,nbatch

            lqz_x = torch.logsumexp(lqz_x, dim=0) - torch.log(
                n_mods_sample
            )  # log_mean_exp
            lpx_z = 0
            for recon_mod in reconstructions[mod]:
                x_recon = reconstructions[mod][recon_mod]
                K, n_batch = x_recon.shape[0], x_recon.shape[1]
                lpx_z_mod = (
                    self.recon_log_probs[recon_mod](x_recon, inputs.data[recon_mod])
                    .view(K, n_batch, -1)
                    .mul(self.rescale_factors[recon_mod])
                    .sum(-1)
                )

                if hasattr(inputs, "masks"):
                    # cancel unavailable modalities
                    lpx_z_mod *= inputs.masks[recon_mod].float()

                lpx_z += lpx_z_mod

            lw = lpx_z + lpz - lqz_x

            if hasattr(inputs, "masks"):
                # cancel unavailable modalities
                lw *= inputs.masks[mod].float()

            lws.append(lw)
            zss.append(z)

        lws = torch.stack(lws)  # (n_modalities, K, n_batch)
        zss = torch.stack(zss)  # (n_modalities, K, n_batch,latent_dim)
        with torch.no_grad():
            grad_wt = (lws - torch.logsumexp(lws, 1, keepdim=True)).exp()
            if zss.requires_grad:  # True except when we are in eval mode
                zss.register_hook(lambda grad: grad_wt.unsqueeze(-1) * grad)

        lws = (grad_wt * lws).sum(0) / n_mods_sample  # mean over modalities

        return ModelOutput(loss=-lws.sum(), metrics=dict(mean_loss_batch=-lws.mean()))

    def iwae(self, qz_xs, zss, reconstructions, inputs):
        lw_mod = []
        for cond_mod in zss:
            lpz = self.prior_dist(*self.pz_params).log_prob(zss[cond_mod]).sum(-1)
            lqz_x = torch.stack(
                [qz_xs[m].log_prob(zss[cond_mod]).sum(-1) for m in qz_xs]
            )
            lqz_x = torch.logsumexp(lqz_x, dim=0) - np.log(lqz_x.size(0))
            lpx_z = 0
            for recon_mod in reconstructions[cond_mod]:
                x_recon = reconstructions[cond_mod][recon_mod]
                K, n_batch = x_recon.shape[0], x_recon.shape[1]
                lpx_z += (
                    self.recon_log_probs[recon_mod](x_recon, inputs.data[recon_mod])
                    .view(K, n_batch, -1)
                    .mul(self.rescale_factors[recon_mod])
                    .sum(-1)
                )
            lw = lpx_z + lpz - lqz_x  # n_samples , n_batch
            lw_mod.append(lw)

        lw = torch.cat(lw_mod, dim=0)  # (n_modalities* K, n_batch)
        lw = torch.logsumexp(lw, dim=0) - np.log(lw.size(0))
        return ModelOutput(loss=-lw.sum(), metrics=dict())

    def encode(
        self,
        inputs: MultimodalBaseDataset,
        cond_mod: Union[list, str] = "all",
        N: int = 1,
        **kwargs,
    ):
        cond_mod = super().encode(inputs, cond_mod, N, **kwargs).cond_mod

        return_mean = kwargs.pop("return_mean", False)
        if all([s in self.encoders.keys() for s in cond_mod]):
            if return_mean:
                emb = torch.stack(
                    [self.encoders[mod](inputs.data[mod]).embedding for mod in cond_mod]
                ).mean(0)
                if N > 1:
                    z = torch.stack([emb] * N)
                else:
                    z = emb

            else:
                # Choose one of the conditioning modalities at random
                mod = np.random.choice(cond_mod)

                output = self.encoders[mod](inputs.data[mod])

                mu, log_var = output.embedding, output.log_covariance
                sigma = self.log_var_to_std(log_var)
                qz_x = self.post_dist(mu, sigma)
                sample_shape = torch.Size([]) if N == 1 else torch.Size([N])
                z = qz_x.rsample(sample_shape)

            flatten = kwargs.pop("flatten", False)
            if flatten:
                z = z.reshape(-1, self.latent_dim)

            return ModelOutput(z=z, one_latent_space=True)

    def compute_joint_nll(
        self, inputs: MultimodalBaseDataset, K: int = 1000, batch_size_K: int = 100
    ):
        """
        Return the estimated negative log-likelihood summed over the inputs.
        The negative log-likelihood is estimated using importance sampling.

        Args:
            inputs : the data to compute the joint likelihood

        """

        self.eval()

        # First compute all the parameters of the joint posterior q(z|x,y)
        post_params = []
        for cond_mod in self.encoders:
            output = self.encoders[cond_mod](inputs.data[cond_mod])
            mu, log_var = output.embedding, output.log_covariance
            sigma = self.log_var_to_std(log_var)
            post_params.append((mu, sigma))

        z_joint = self.encode(inputs, N=K).z
        z_joint = z_joint.permute(1, 0, 2)
        n_data, _, latent_dim = z_joint.shape

        # Then iter on each datapoint to compute the iwae estimate of ln(p(x))
        ll = 0
        for i in range(n_data):
            start_idx = 0
            stop_idx = min(start_idx + batch_size_K, K)
            lnpxs = []
            while start_idx < stop_idx:
                latents = z_joint[i][start_idx:stop_idx]

                # Compute p(x_m|z) for z in latents and for each modality m
                lpx_zs = 0  # ln(p(x,y|z))
                for mod in inputs.data:
                    decoder = self.decoders[mod]
                    recon = decoder(latents)[
                        "reconstruction"
                    ]  # (batch_size_K, nb_channels, w, h)
                    x_m = inputs.data[mod][i]  # (nb_channels, w, h)

                    lpx_zs += (
                        self.recon_log_probs[mod](recon, x_m)
                        .reshape(recon.size(0), -1)
                        .sum(-1)
                    )

                # Compute ln(p(z))
                prior = self.prior_dist(*self.pz_params)
                lpz = prior.log_prob(latents).sum(dim=-1)

                # Compute posteriors -ln(q(z|x,y))
                qz_xs = [self.post_dist(p[0][i], p[1][i]) for p in post_params]
                lqz_xy = torch.logsumexp(
                    torch.stack([q.log_prob(latents).sum(-1) for q in qz_xs]), dim=0
                ) - np.log(self.n_modalities)

                ln_px = torch.logsumexp(lpx_zs + lpz - lqz_xy, dim=0)
                lnpxs.append(ln_px)

                # next batch
                start_idx += batch_size_K
                stop_idx = min(stop_idx + batch_size_K, K)

            ll += torch.logsumexp(torch.Tensor(lnpxs), dim=0) - np.log(K)

        return -ll

    @torch.no_grad()
    def compute_joint_nll_paper(
        self, inputs: MultimodalBaseDataset, K: int = 1000, batch_size_K: int = 10
    ):
        """Computes the joint likelihood like in the original dataset, using all Mixture of experts
        samples and modality rescaling."""

        self.eval()

        lws = []
        nb_computed_samples = 0
        while nb_computed_samples < K:
            n_samples = min(batch_size_K, K - nb_computed_samples)
            nb_computed_samples += n_samples
            # Compute a iwae likelihood estimate using n_samples
            output = self.forward(
                inputs, compute_loss=False, K=n_samples, detailed_output=True
            )
            lw = self.iwae(output.qz_xs, output.zss, output.recon, inputs).loss
            lws.append(lw + np.log(n_samples * self.n_modalities))

        ll = torch.logsumexp(torch.stack(lws), dim=0) - np.log(
            nb_computed_samples * self.n_modalities
        )  # n_batch
        return -ll

    def generate_from_prior(self, n_samples, **kwargs):
        sample_shape = [n_samples] if n_samples > 1 else []
        z = self.prior_dist(*self.pz_params).rsample(sample_shape).to(self.device)
        return ModelOutput(z=z.squeeze(), one_latent_space=True)



from pydantic.dataclasses import dataclass

from ..base.evaluator_config import EvaluatorConfig


@dataclass
class CoherenceEvaluatorConfig(EvaluatorConfig):
    """

    Config class for the evaluation of the coherences module.

    Args:
        batch_size (int) : The batch size to use in the evaluation.
        wandb_path (str) : The user can provide the path of the wandb run with a
            format 'entity/projet_name/run_id' where the metrics should be logged.
            For an existing run (the training run), the info can be found in the training dir (in wandb_info.json)
            at the end of training (if wandb was used) or on the hugging_face webpage of the run.
            Otherwise the user can create a new wandb run and get the path with :

                .. code-block:: python

                    >>> import wandb
                    >>> run = wandb.init(entity = your_entity, project=your_project)
                    >>> wandb_path = run.path


            If None are provided, the metrics are not logged on wandb.
            Default to None.
        include_recon (bool) : If True, we include the reconstructions in the mean conditional generations
            coherences. Default to False.
        nb_samples_for_joint (int): How many samples to use to compute joint coherence. Default to 10000.


    """

    include_recon: bool = False
    nb_samples_for_joint: int = 10000



from ..base import BaseTrainerConfig


class TwoStepsTrainerConfig(BaseTrainerConfig):
    """
    Configuration for a specific trainer that handles the training of the joint VAE models.

    Parameters:

        output_dir (str): The directory where model checkpoints, configs and final
            model will be stored. Default: None.
        per_device_train_batch_size (int): The number of training samples per batch and per device.
            Default 64
        per_device_eval_batch_size (int): The number of evaluation samples per batch and per device.
            Default 64
        num_epochs (int): The maximal number of epochs for training. Default: 100
        train_dataloader_num_workers (int): Number of subprocesses to use for train data loading.
            0 means that the data will be loaded in the main process. Default: 0
        eval_dataloader_num_workers (int): Number of subprocesses to use for evaluation data
            loading. 0 means that the data will be loaded in the main process. Default: 0
        optimizer_cls (str): The name of the `torch.optim.Optimizer` used for
            training. Default: :class:`~torch.optim.Adam`.
        optimizer_params (dict): A dict containing the parameters to use for the
            `torch.optim.Optimizer`. If None, uses the default parameters. Default: None.
        scheduler_cls (str): The name of the `torch.optim.lr_scheduler` used for
            training. If None, no scheduler is used. Default None.
        scheduler_params (dict): A dict containing the parameters to use for the
            `torch.optim.le_scheduler`. If None, uses the default parameters. Default: None.
        learning_rate (int): The learning rate applied to the `Optimizer`. Default: 1e-4
        steps_saving (int): A model checkpoint will be saved every `steps_saving` epoch.
            Default: None
        steps_predict (int): A prediction using the best model will be run every `steps_predict`
            epoch. Default: None
        keep_best_on_train (bool): Whether to keep the best model on the train set. Default: False
        seed (int): The random seed for reproducibility
        no_cuda (bool): Disable `cuda` training. Default: False
        world_size (int): The total number of process to run. Default: -1
        local_rank (int): The rank of the node for distributed training. Default: -1
        rank (int): The rank of the process for distributed training. Default: -1
        dist_backend (str): The distributed backend to use. Default: 'nccl'
        master_addr (str): The master address for distributed training. Default: 'localhost'
        master_port (str): The master port for distributed training. Default: '12345'
        start_keep_best (int): The epoch to start keeping track of the best loss. Before that the model
            is automatically saved.
    """

    name_trainer = "TwoStepsTrainerConfig"



from pydantic.dataclasses import dataclass

from ..base.base_config import BaseMultiVAEConfig


@dataclass
class MVTCAEConfig(BaseMultiVAEConfig):
    """
    This is the base config class for the MVTCAE model from
    'Multi-View Representation Learning via Total Correlation Objective' Neurips 2021.
    The code is based on the original implementation that can be found here :
    https://github.com/gr8joo/MVTCAE/blob/master/run_epochs.py

    Args:
        n_modalities (int): The number of modalities. Default: None.
        latent_dim (int): The dimension of the latent space. Default: None.
        input_dims (dict[str,tuple]) : The modalities'names (str) and input shapes (tuple).
        uses_likelihood_rescaling (bool): To mitigate modality collapse, it is possible to use likelihood rescaling.
            (see : https://proceedings.mlr.press/v162/javaloy22a.html).
            The inputs_dim must be provided to compute the likelihoods rescalings. It is used in a number of models
            which is why we include it here. Default to False.
        decoders_dist (Dict[str, Union[function, str]]). The decoder distributions to use per modality.
            Per modality, you can provide a string in ['normal','bernoulli','laplace']. If None is provided,
            a normal distribution is used for each modality.
        decoder_dist_params (Dict[str,dict]) : Parameters for the output decoder distributions, for
            computing the log-probability.
            For instance, with normal or laplace distribution, you can pass the scale in this dictionary.
            ex :  {'mod1' : {scale : 0.75}}
        alpha (float) : The parameter that ponderates the total correlation ratio in the loss.
        beta (float) : The parameter that weights the sum of all KLs
    """

    alpha: float = 0.1
    beta: float = 2.5



import os
from copy import deepcopy

import numpy as np
import pytest
import torch
from encoders import Encoder_test
from pytest import fixture
from pythae.models.base import BaseAEConfig
from pythae.models.base.base_utils import ModelOutput
from pythae.models.normalizing_flows import IAF, IAFConfig

from multivae.data.datasets import MultimodalBaseDataset
from multivae.models.auto_model.auto_model import AutoModel
from multivae.models.dcca import DCCA, DCCAConfig
from multivae.models.jnf_dcca import JNFDcca, JNFDccaConfig
from multivae.models.nn.default_architectures import BaseDictEncoders, Decoder_AE_MLP
from multivae.trainers.add_dcca_trainer import AddDccaTrainer, AddDccaTrainerConfig


class TestDcca:
    @fixture(
        params=[
            dict(n_modalities=2, embedding_dim=5, use_all_singular_values=False),
            dict(n_modalities=5, embedding_dim=10, use_all_singular_values=False),
        ]
    )
    def inputs(self, request):
        n_mod = request.param["n_modalities"]
        embed_dim = request.param["embedding_dim"]
        networks = BaseDictEncoders(
            {"mod" + str(k): (k,) for k in range(1, n_mod + 1)}, embed_dim
        )

        data = {"mod" + str(k): torch.ones((10, k)) for k in range(1, n_mod + 1)}
        return networks, DCCAConfig(**request.param), MultimodalBaseDataset(data)

    def test(self, inputs):
        networks, config, data = inputs
        model = DCCA(config, networks)

        assert model.latent_dim == config.embedding_dim
        assert model.use_all_singular_values == config.use_all_singular_values

        output = model(data)
        assert hasattr(output, "loss")

    def test_raises_error_forward(self, inputs):
        networks, config, data = inputs
        model = DCCA(config, networks)

        with pytest.raises(AttributeError):
            model(MultimodalBaseDataset({"unknown_modality": 10}))

    def test_set_networks(self, inputs):
        with pytest.raises(AssertionError):
            _ = DCCA(inputs[1], {"mod1": AutoModel()})

        inputs[0]["mod1"].latent_dim = inputs[1].embedding_dim + 1

        with pytest.raises(AttributeError):
            _ = DCCA(inputs[1], inputs[0])


class TestJNFDcca:
    @fixture
    def dataset(self):
        # Create simple small dataset with 2 modalities
        data = dict(
            mod1=torch.rand((200, 2)),
            mod2=torch.rand((200, 3)),
            mod3=torch.rand((200, 4)),
        )
        labels = np.random.randint(2, size=200)
        dataset = MultimodalBaseDataset(data, labels)
        return dataset

    @fixture
    def custom_architectures(self):
        # Create custom instances for the dcca_networks and decoders
        config_dcca_1 = BaseAEConfig(input_dim=(2,), latent_dim=2)
        config_dcca_2 = BaseAEConfig(input_dim=(3,), latent_dim=2)
        config_dcca_3 = BaseAEConfig(input_dim=(4,), latent_dim=2)

        config1 = BaseAEConfig(input_dim=(2,), latent_dim=5)
        config2 = BaseAEConfig(input_dim=(3,), latent_dim=5)
        config3 = BaseAEConfig(input_dim=(4,), latent_dim=5)

        dcca_networks = dict(
            mod1=Encoder_test(config_dcca_1),
            mod2=Encoder_test(config_dcca_2),
            mod3=Encoder_test(config_dcca_3),
        )

        decoders = dict(
            mod1=Decoder_AE_MLP(config1),
            mod2=Decoder_AE_MLP(config2),
            mod3=Decoder_AE_MLP(config3),
        )

        flows = dict(
            mod1=IAF(IAFConfig(input_dim=(5,))),
            mod2=IAF(IAFConfig(input_dim=(5,))),
            mod3=IAF(IAFConfig(input_dim=(5,))),
        )
        return dict(
            dcca_networks=dcca_networks,
            decoders=decoders,
            flows=flows,
        )

    @fixture(params=[False])
    def model_config(self, request):
        model_config = JNFDccaConfig(
            n_modalities=3,
            latent_dim=5,
            input_dims=dict(mod1=(2,), mod2=(3,), mod3=(4,)),
            use_all_singular_values=request.param,
            embedding_dcca_dim=2,
            nb_epochs_dcca=2,
            warmup=2,
        )

        return model_config

    @fixture(params=[True, False])
    def model(self, custom_architectures, model_config, request):
        custom = request.param
        if custom:
            model = JNFDcca(model_config, **custom_architectures)
        else:
            model = JNFDcca(model_config)

        return model

    @fixture
    def training_config(self, tmpdir):
        tmpdir.mkdir("dummy_folder")
        dir_path = os.path.join(tmpdir, "dummy_folder")
        return AddDccaTrainerConfig(
            num_epochs=5,
            steps_saving=2,
            learning_rate=1e-3,
            optimizer_cls="AdamW",
            optimizer_params={"betas": (0.91, 0.995)},
            output_dir=dir_path,
            learning_rate_dcca=1e-4,
        )

    @fixture
    def trainer(self, model, training_config, dataset):
        trainer = AddDccaTrainer(
            model=model,
            train_dataset=dataset,
            eval_dataset=dataset,
            training_config=training_config,
        )

        trainer.prepare_training()

        return trainer

    def test_model_forward(self, model, dataset, model_config):
        assert hasattr(model, "dcca_networks")
        assert model.warmup == model_config.warmup
        assert model.nb_epochs_dcca == model_config.nb_epochs_dcca

        # Test forward method during dcca training
        output = model(dataset, epoch=1)
        assert not hasattr(output, "recon_loss")
        assert not hasattr(output, "KLD")
        loss = output.loss
        assert type(loss) == torch.Tensor
        assert loss.size() == torch.Size([])
        assert loss.requires_grad

        # Test forward method during joint vae training
        output = model(dataset, epoch=model_config.nb_epochs_dcca + 1)
        assert hasattr(output, "recon_loss")
        assert hasattr(output, "KLD")
        loss = output.loss
        assert type(loss) == torch.Tensor
        assert loss.size() == torch.Size([])
        # assert loss.requires_grad

        # Test forward method during flows training
        output = model(
            dataset, epoch=model_config.nb_epochs_dcca + model_config.warmup + 2
        )
        assert hasattr(output, "ljm")
        loss = output.loss
        assert type(loss) == torch.Tensor
        assert loss.size() == torch.Size([])
        assert loss.requires_grad

        # Try encoding and prediction
        outputs = model.encode(dataset)
        assert outputs.one_latent_space
        embeddings = outputs.z
        assert isinstance(outputs, ModelOutput)
        assert embeddings.shape == (200, model_config.latent_dim)
        embeddings = model.encode(dataset, N=2).z
        assert embeddings.shape == (2, 200, model_config.latent_dim)
        embeddings = model.encode(dataset, cond_mod=["mod1"]).z
        assert embeddings.shape == (200, model_config.latent_dim)
        embeddings = model.encode(dataset, cond_mod="mod2", N=10).z
        assert embeddings.shape == (10, 200, model_config.latent_dim)
        embeddings = model.encode(dataset, cond_mod=["mod2", "mod1"], mcmc_steps=2).z
        assert embeddings.shape == (200, model_config.latent_dim)

        Y = model.predict(dataset, cond_mod="mod1")
        assert isinstance(Y, ModelOutput)
        assert Y.mod1.shape == (200, 2)
        assert Y.mod2.shape == (200, 3)

        Y = model.predict(dataset, cond_mod="mod1", N=10)
        assert isinstance(Y, ModelOutput)
        assert Y.mod1.shape == (10, 200, 2)
        assert Y.mod2.shape == (10, 200, 3)

        Y = model.predict(dataset, cond_mod="mod1", N=10, flatten=True)
        assert isinstance(Y, ModelOutput)
        assert Y.mod1.shape == (200 * 10, 2)
        assert Y.mod2.shape == (200 * 10, 3)

    @pytest.mark.slow
    def test_train_step(self, trainer):
        start_model_state_dict = deepcopy(trainer.model.state_dict())
        start_optimizer = trainer.optimizer
        _ = trainer.train_step(epoch=1)

        step_1_model_state_dict = deepcopy(trainer.model.state_dict())

        # check that weights were updated
        assert not all(
            [
                torch.equal(start_model_state_dict[key], step_1_model_state_dict[key])
                for key in start_model_state_dict.keys()
            ]
        )
        assert trainer.optimizer == start_optimizer
        assert trainer.training_config.learning_rate == 1e-4
        _ = trainer.prepare_train_step(trainer.model.nb_epochs_dcca + 1, None, None)
        _ = trainer.train_step(epoch=trainer.model.nb_epochs_dcca + 1)
        assert trainer.training_config.learning_rate == 1e-3
        step_2_model_state_dict = deepcopy(trainer.model.state_dict())

        assert not all(
            [
                torch.equal(step_1_model_state_dict[key], step_2_model_state_dict[key])
                for key in step_1_model_state_dict.keys()
            ]
        )
        assert trainer.optimizer != start_optimizer

    @pytest.mark.slow
    def test_eval_step(self, trainer):
        start_model_state_dict = deepcopy(trainer.model.state_dict())

        _ = trainer.eval_step(epoch=1)

        step_1_model_state_dict = deepcopy(trainer.model.state_dict())

        # check that weights were not updated
        assert all(
            [
                torch.equal(start_model_state_dict[key], step_1_model_state_dict[key])
                for key in start_model_state_dict.keys()
            ]
        )

    @pytest.mark.slow
    def test_main_train_loop(self, trainer):
        start_model_state_dict = deepcopy(trainer.model.state_dict())

        trainer.train()

        step_1_model_state_dict = deepcopy(trainer.model.state_dict())

        # check that weights were updated
        assert not all(
            [
                torch.equal(start_model_state_dict[key], step_1_model_state_dict[key])
                for key in start_model_state_dict.keys()
            ]
        )

    @pytest.mark.slow
    def test_checkpoint_saving(self, model, trainer, training_config):
        assert hasattr(trainer.model, "dcca_networks")
        dir_path = training_config.output_dir

        # Make a training step
        step_1_loss = trainer.train_step(epoch=1)

        model = deepcopy(trainer.model)

        assert hasattr(trainer.model, "dcca_networks")

        optimizer = deepcopy(trainer.optimizer)

        trainer.save_checkpoint(dir_path=dir_path, epoch=0, model=model)

        checkpoint_dir = os.path.join(dir_path, "checkpoint_epoch_0")

        assert os.path.isdir(checkpoint_dir)

        files_list = os.listdir(checkpoint_dir)

        assert set(["model.pt", "optimizer.pt", "training_config.json"]).issubset(
            set(files_list)
        )

        # check pickled custom architectures
        for archi in model.model_config.custom_architectures:
            assert archi + ".pkl" in files_list

        model_rec_state_dict = torch.load(os.path.join(checkpoint_dir, "model.pt"))[
            "model_state_dict"
        ]

        assert all(
            [
                torch.equal(
                    model_rec_state_dict[key].cpu(), model.state_dict()[key].cpu()
                )
                for key in model.state_dict().keys()
            ]
        )

        # check reload full model
        model_rec = AutoModel.load_from_folder(os.path.join(checkpoint_dir))

        assert all(
            [
                torch.equal(
                    model_rec.state_dict()[key].cpu(), model.state_dict()[key].cpu()
                )
                for key in model.state_dict().keys()
            ]
        )

        assert type(model_rec.encoders.cpu()) == type(model.encoders.cpu())
        assert type(model_rec.decoders.cpu()) == type(model.decoders.cpu())

        optim_rec_state_dict = torch.load(os.path.join(checkpoint_dir, "optimizer.pt"))

        assert all(
            [
                dict_rec == dict_optimizer
                for (dict_rec, dict_optimizer) in zip(
                    optim_rec_state_dict["param_groups"],
                    optimizer.state_dict()["param_groups"],
                )
            ]
        )

        assert all(
            [
                dict_rec == dict_optimizer
                for (dict_rec, dict_optimizer) in zip(
                    optim_rec_state_dict["state"], optimizer.state_dict()["state"]
                )
            ]
        )

    @pytest.mark.slow
    def test_checkpoint_saving_during_training(self, model, trainer, training_config):
        #
        target_saving_epoch = training_config.steps_saving

        dir_path = training_config.output_dir

        model = deepcopy(trainer.model)

        trainer.train()

        training_dir = os.path.join(
            dir_path, f"JNFDcca_training_{trainer._training_signature}"
        )
        assert os.path.isdir(training_dir)

        checkpoint_dir = os.path.join(
            training_dir, f"checkpoint_epoch_{target_saving_epoch}"
        )

        assert os.path.isdir(checkpoint_dir)

        files_list = os.listdir(checkpoint_dir)

        # check files
        assert set(["model.pt", "optimizer.pt", "training_config.json"]).issubset(
            set(files_list)
        )

        # check pickled custom architectures
        for archi in model.model_config.custom_architectures:
            assert archi + ".pkl" in files_list

        model_rec_state_dict = torch.load(os.path.join(checkpoint_dir, "model.pt"))[
            "model_state_dict"
        ]

        assert not all(
            [
                torch.equal(model_rec_state_dict[key], model.state_dict()[key])
                for key in model.state_dict().keys()
            ]
        )

    @pytest.mark.slow
    def test_final_model_saving(self, model, trainer, training_config):
        dir_path = training_config.output_dir

        trainer.train()

        model = deepcopy(trainer._best_model)

        training_dir = os.path.join(
            dir_path, f"JNFDcca_training_{trainer._training_signature}"
        )
        assert os.path.isdir(training_dir)

        final_dir = os.path.join(training_dir, f"final_model")
        assert os.path.isdir(final_dir)

        files_list = os.listdir(final_dir)

        assert set(["model.pt", "model_config.json", "training_config.json"]).issubset(
            set(files_list)
        )

        # check reload full model
        model_rec = AutoModel.load_from_folder(os.path.join(final_dir))

        assert all(
            [
                torch.equal(
                    model_rec.state_dict()[key].cpu(), model.state_dict()[key].cpu()
                )
                for key in model.state_dict().keys()
            ]
        )

        assert type(model_rec.encoders.cpu()) == type(model.encoders.cpu())
        assert type(model_rec.decoders.cpu()) == type(model.decoders.cpu())
        assert type(model_rec.DCCA_module.networks.cpu()) == type(
            model.DCCA_module.networks.cpu()
        )

    def test_compute_nll(self, model, dataset):
        nll = model.compute_joint_nll(dataset, K=10, batch_size_K=2)
        assert nll >= 0
        assert type(nll) == torch.Tensor
        assert nll.size() == torch.Size([])



import numpy as np
import pytest
import torch
from pythae.models.base import BaseAEConfig

from multivae.models.nn.mmnist import (
    Decoder_ResNet_AE_MMNIST,
    DecoderConvMMNIST,
    Encoder_ResNet_VAE_MMNIST,
    EncoderConvMMNIST,
)
from multivae.models.nn.svhn import Decoder_VAE_SVHN, Encoder_VAE_SVHN
from multivae.models.nn.cub import CubTextEncoder, CubTextDecoderMLP

device = "cuda" if torch.cuda.is_available() else "cpu"


#### MMNIST configs ####
@pytest.fixture(
    params=[
        BaseAEConfig(input_dim=(3, 28, 28), latent_dim=10),
        BaseAEConfig(input_dim=(3, 28, 28), latent_dim=5),
    ]
)
def ae_mmnist_config(request):
    return request.param


@pytest.fixture()
def mmnist_like_data():
    return torch.rand(3, 3, 28, 28).to(device)


#### CIFAR configs ####
@pytest.fixture(
    params=[
        BaseAEConfig(input_dim=(3, 32, 32), latent_dim=10),
        BaseAEConfig(input_dim=(3, 32, 32), latent_dim=5),
    ]
)
def ae_svhn_config(request):
    return request.param


@pytest.fixture()
def svhn_like_data():
    return torch.rand(3, 3, 32, 32).to(device)


#### CUB configs ####
@pytest.fixture(
    params=[
        BaseAEConfig(input_dim=(12, 356), latent_dim=10),
        BaseAEConfig(input_dim=(18, 14), latent_dim=5),
    ]
)
def ae_cub_config(request):
    return request.param


@pytest.fixture()
def cub_like_data(ae_cub_config):
    return dict(
        tokens=torch.randint(0, ae_cub_config.input_dim[1], (3, ae_cub_config.input_dim[0])).to(device),
        padding_mask=torch.randint(0, 1, (3, ae_cub_config.input_dim[0])).to(device).type(torch.float)
    )


class TestMMNISTNets:
    def test_forward(self, ae_mmnist_config, mmnist_like_data):
        encoder = EncoderConvMMNIST(ae_mmnist_config).to(device)
        decoder = DecoderConvMMNIST(ae_mmnist_config).to(device)

        enc_out = encoder(mmnist_like_data)

        assert enc_out.embedding.shape == (
            mmnist_like_data.shape[0],
            ae_mmnist_config.latent_dim,
        )
        assert enc_out.log_covariance.shape == (
            mmnist_like_data.shape[0],
            ae_mmnist_config.latent_dim,
        )

        dec_out = decoder(enc_out.embedding)

        assert dec_out.reconstruction.shape == mmnist_like_data.shape

        encoder = Encoder_ResNet_VAE_MMNIST(ae_mmnist_config).to(device)
        decoder = Decoder_ResNet_AE_MMNIST(ae_mmnist_config).to(device)

        enc_out = encoder(mmnist_like_data)

        assert enc_out.embedding.shape == (
            mmnist_like_data.shape[0],
            ae_mmnist_config.latent_dim,
        )
        assert enc_out.log_covariance.shape == (
            mmnist_like_data.shape[0],
            ae_mmnist_config.latent_dim,
        )

        dec_out = decoder(enc_out.embedding)

        assert dec_out.reconstruction.shape == mmnist_like_data.shape


class TestSVHNNets:
    def test_forward(self, ae_svhn_config, svhn_like_data):
        encoder = Encoder_VAE_SVHN(ae_svhn_config).to(device)
        decoder = Decoder_VAE_SVHN(ae_svhn_config).to(device)

        enc_out = encoder(svhn_like_data)

        assert enc_out.embedding.shape == (
            svhn_like_data.shape[0],
            ae_svhn_config.latent_dim,
        )
        assert enc_out.log_covariance.shape == (
            svhn_like_data.shape[0],
            ae_svhn_config.latent_dim,
        )

        dec_out = decoder(enc_out.embedding)

        assert dec_out.reconstruction.shape == svhn_like_data.shape


class TestCUBNets:
    def test_forward(self, ae_cub_config, cub_like_data):
        encoder = CubTextEncoder(
            ae_cub_config,
            max_sentence_length=ae_cub_config.input_dim[0],
            ntokens=ae_cub_config.input_dim[1],
            embed_size=512,
            ff_size=128
        ).to(device)
        decoder = CubTextDecoderMLP(ae_cub_config).to(device)

        enc_out = encoder(cub_like_data)

        assert enc_out.embedding.shape == (
            cub_like_data['tokens'].shape[0],
            ae_cub_config.latent_dim,
        )
        assert enc_out.log_covariance.shape == (
            cub_like_data['tokens'].shape[0],
            ae_cub_config.latent_dim,
        )

        dec_out = decoder(enc_out.embedding)

        assert dec_out.reconstruction.shape == cub_like_data["tokens"].shape + (ae_cub_config.input_dim[1],)



"""In this section, you will find all the trainers that are currently implemented in `multivae` library
"""

from .add_dcca_trainer import AddDccaTrainer, AddDccaTrainerConfig
from .base import BaseTrainer, BaseTrainerConfig
from .jnf import TwoStepsTrainer, TwoStepsTrainerConfig

__all__ = [
    "BaseTrainer",
    "BaseTrainerConfig",
    "TwoStepsTrainer",
    "TwoStepsTrainerConfig",
    "AddDccaTrainer",
    "AddDccaTrainerConfig",
]



import torch
from pythae.models.base.base_utils import ModelOutput
from pythae.models.nn.base_architectures import BaseDecoder, BaseEncoder
from torch import nn


class Encoder_VAE_SVHN(BaseEncoder):
    def __init__(self, args):
        BaseEncoder.__init__(self)
        self.input_dim = args.input_dim
        self.latent_dim = args.latent_dim
        self.n_channels = args.input_dim[0]
        self.fBase = 32

        self.enc = nn.Sequential(
            # input size: 3 x 32 x 32
            nn.Conv2d(self.n_channels, self.fBase, 4, 2, 1, bias=True),
            nn.ReLU(True),
            # size: (fBase) x 16 x 16
            nn.Conv2d(self.fBase, self.fBase * 2, 4, 2, 1, bias=True),
            nn.ReLU(True),
            # size: (fBase * 2) x 8 x 8
            nn.Conv2d(self.fBase * 2, self.fBase * 4, 4, 2, 1, bias=True),
            nn.ReLU(True),
            # size: (fBase * 4) x 4 x 4
        )
        self.c1 = nn.Conv2d(self.fBase * 4, self.latent_dim, 4, 2, 0)
        self.c2 = nn.Conv2d(self.fBase * 4, self.latent_dim, 4, 2, 0)
        # c1, c2 size: latent_dim x 1 x 1

    def forward(self, x: torch.Tensor):
        e = self.enc(x)
        mu = self.c1(e).squeeze()
        lv = self.c2(e).squeeze()
        output = ModelOutput(embedding=mu, log_covariance=lv)
        return output


class Decoder_VAE_SVHN(BaseDecoder):
    def __init__(self, args):
        BaseDecoder.__init__(self)
        self.latent_dim = args.latent_dim
        self.fBase = 32
        self.nb_channels = args.input_dim[0]

        self.dec = nn.Sequential(
            nn.ConvTranspose2d(self.latent_dim, self.fBase * 4, 4, 1, 0, bias=True),
            nn.ReLU(True),
            # size: (fBase * 4) x 4 x 4
            nn.ConvTranspose2d(self.fBase * 4, self.fBase * 2, 4, 2, 1, bias=True),
            nn.ReLU(True),
            # size: (fBase * 2) x 8 x 8
            nn.ConvTranspose2d(self.fBase * 2, self.fBase, 4, 2, 1, bias=True),
            nn.ReLU(True),
            # size: (fBase) x 16 x 16
            nn.ConvTranspose2d(self.fBase, self.nb_channels, 4, 2, 1, bias=True),
            nn.Sigmoid()
            # Output size: 3 x 32 x 32
        )

    def forward(self, z: torch.Tensor):
        z = z.unsqueeze(-1).unsqueeze(-1)  # fit deconv layers
        out = self.dec(z.view(-1, *z.size()[-3:]))
        out = out.view(*z.size()[:-3], *out.size()[1:])
        output = ModelOutput(reconstruction=out)
        return output



import os
from copy import deepcopy

import numpy as np
import pytest
import torch
from PIL import Image
from tests_data.classifiers import MNIST_Classifier, SVHN_Classifier
from torch import nn

from multivae.data.datasets.base import MultimodalBaseDataset
from multivae.metrics.base import Evaluator, EvaluatorConfig
from multivae.metrics.coherences import CoherenceEvaluator, CoherenceEvaluatorConfig
from multivae.metrics.likelihoods import (
    LikelihoodsEvaluator,
    LikelihoodsEvaluatorConfig,
)
from multivae.models import JMVAE, JMVAEConfig, MoPoE, MoPoEConfig
from multivae.samplers import GaussianMixtureSampler, GaussianMixtureSamplerConfig


@pytest.fixture
def jmvae_model():
    return JMVAE(
        JMVAEConfig(
            n_modalities=2, input_dims=dict(mnist=(1, 28, 28), svhn=(3, 32, 32))
        )
    )


@pytest.fixture
def dataset():
    return MultimodalBaseDataset(
        {"mnist": torch.randn(30, 1, 28, 28), "svhn": torch.randn(30, 3, 32, 32)},
        labels=torch.ones(30),
    )


@pytest.fixture
def output_logger_file(tmpdir):
    os.mkdir(os.path.join(tmpdir, "logger_metrics"))
    return os.path.join(tmpdir, "logger_metrics")


class TestBaseMetric:
    @pytest.fixture(params=[10, 23])
    def batch_size(self, request):
        return request.param

    def test_evaluator_config(self, batch_size):
        config = EvaluatorConfig(batch_size=batch_size)
        assert config.batch_size == batch_size

    def test_evaluator_class(
        self, batch_size, jmvae_model, output_logger_file, dataset
    ):
        config = EvaluatorConfig(batch_size=batch_size)
        evaluator = Evaluator(
            jmvae_model, dataset, output=output_logger_file, eval_config=config
        )

        assert evaluator.n_data == len(dataset.data["mnist"])
        assert os.path.exists(os.path.join(output_logger_file, "metrics.log"))

        assert (
            next(iter(evaluator.test_loader)).data["mnist"].shape
            == (batch_size,) + dataset.data["mnist"].shape[1:]
        )


class TestCoherences:
    @pytest.fixture(params=[[True, 21], [False, 3]])
    def config_params(self, request):
        return {
            "include_recon": request.param[0],
            "nb_samples_for_joint": request.param[1],
        }

    @pytest.fixture
    def classifiers(self):
        return dict(mnist=MNIST_Classifier(), svhn=SVHN_Classifier())

    @pytest.fixture(params=[True, False])
    def sampler(self, jmvae_model, dataset, request):
        if not request.param:
            return None
        else:
            config = GaussianMixtureSamplerConfig(n_components=3)
            sampler = GaussianMixtureSampler(jmvae_model, config)
            sampler.fit(dataset)
            return sampler

    def test_coherence_config(self, config_params):
        config = CoherenceEvaluatorConfig(
            include_recon=config_params["include_recon"],
            nb_samples_for_joint=config_params["nb_samples_for_joint"],
        )

        assert config.include_recon == config_params["include_recon"]
        assert config.nb_samples_for_joint == config_params["nb_samples_for_joint"]

    def test_cross_coherence_compute(
        self, jmvae_model, config_params, classifiers, output_logger_file, dataset
    ):
        config = CoherenceEvaluatorConfig(
            include_recon=config_params["include_recon"],
            nb_samples_for_joint=config_params["nb_samples_for_joint"],
        )

        evaluator = CoherenceEvaluator(
            model=jmvae_model,
            classifiers=classifiers,
            output=output_logger_file,
            test_dataset=dataset,
            eval_config=config,
        )

        cross_coherences = evaluator.cross_coherences()
        assert all([0 <= cc_score[0] <= 1 for cc_score in cross_coherences])

    def test_joint_coherence_compute(
        self,
        jmvae_model,
        config_params,
        classifiers,
        output_logger_file,
        dataset,
        sampler,
    ):
        config = CoherenceEvaluatorConfig(
            include_recon=config_params["include_recon"],
            nb_samples_for_joint=config_params["nb_samples_for_joint"],
        )

        evaluator = CoherenceEvaluator(
            model=jmvae_model,
            classifiers=classifiers,
            output=output_logger_file,
            test_dataset=dataset,
            eval_config=config,
            sampler=sampler,
        )

        joint_coherence = evaluator.joint_coherence()
        assert 0 <= joint_coherence <= 1

    def test_eval(
        self, jmvae_model, config_params, classifiers, output_logger_file, dataset
    ):
        config = CoherenceEvaluatorConfig(
            include_recon=config_params["include_recon"],
            nb_samples_for_joint=config_params["nb_samples_for_joint"],
        )

        evaluator = CoherenceEvaluator(
            model=jmvae_model,
            classifiers=classifiers,
            output=output_logger_file,
            test_dataset=dataset,
            eval_config=config,
        )

        metrics = evaluator.eval()
        assert all(
            [
                metric in metrics.keys()
                for metric in ["mean_coherence_1", "joint_coherence_prior"]
            ]
        )


class TestLikelihoods:
    @pytest.fixture(params=[[2, 21], [16, 3]])
    def config_params(self, request):
        return {"num_samples": request.param[0], "batch_size": request.param[1]}

    @pytest.fixture
    def mopoe_model(self):
        return MoPoE(
            MoPoEConfig(
                n_modalities=2, input_dims=dict(mnist=(1, 28, 28), svhn=(3, 32, 32))
            )
        )

    def test_likelihood_config(self, config_params):
        config = LikelihoodsEvaluatorConfig(
            num_samples=config_params["num_samples"],
            batch_size=config_params["batch_size"],
        )

        assert config.num_samples == config_params["num_samples"]
        assert config.batch_size == config_params["batch_size"]

    def test_joint_nll(self, jmvae_model, config_params, output_logger_file, dataset):
        config = LikelihoodsEvaluatorConfig(
            num_samples=config_params["num_samples"],
            batch_size=config_params["batch_size"],
        )

        evaluator = LikelihoodsEvaluator(
            model=jmvae_model,
            output=output_logger_file,
            test_dataset=dataset,
            eval_config=config,
        )

        joint_nll = evaluator.joint_nll()

        assert joint_nll > 0

    def test_joint_nll_from_subset(
        self, mopoe_model, jmvae_model, config_params, output_logger_file, dataset
    ):
        config = LikelihoodsEvaluatorConfig(
            num_samples=config_params["num_samples"],
            batch_size=config_params["batch_size"],
        )

        evaluator = LikelihoodsEvaluator(
            model=mopoe_model,
            output=output_logger_file,
            test_dataset=dataset,
            eval_config=config,
        )

        joint_nll_from_sub = evaluator.joint_nll_from_subset(["mnist", "svhn"])
        assert joint_nll_from_sub > 0

        evaluator = LikelihoodsEvaluator(
            model=jmvae_model,
            output=output_logger_file,
            test_dataset=dataset,
            eval_config=config,
        )

        joint_nll_from_sub = evaluator.joint_nll_from_subset(["mnist", "svhn"])
        assert joint_nll_from_sub is None

    def test_eval(self, jmvae_model, config_params, output_logger_file, dataset):
        config = LikelihoodsEvaluatorConfig(
            num_samples=config_params["num_samples"],
            batch_size=config_params["batch_size"],
        )

        evaluator = LikelihoodsEvaluator(
            model=jmvae_model,
            output=output_logger_file,
            test_dataset=dataset,
            eval_config=config,
        )

        metrics = evaluator.eval()

        assert all([metric in metrics.keys() for metric in ["joint_likelihood"]])


import tempfile

from multivae.metrics import Visualization, VisualizationConfig
from multivae.models.base import ModelOutput


class Test_Visualization:
    def test(self, jmvae_model, dataset):
        tmpdir = tempfile.mkdtemp()

        module = Visualization(model=jmvae_model, output=tmpdir, test_dataset=dataset)

        output = module.eval()
        assert isinstance(output, ModelOutput)
        assert hasattr(output, "unconditional_generation")
        assert os.path.exists(os.path.join(tmpdir, "unconditional.png"))

        output = module.conditional_samples_subset(["mnist"])
        assert isinstance(output, Image.Image)


from multivae.metrics import Clustering, ClusteringConfig


class Test_clustering:
    def test(self, jmvae_model, dataset):
        tmpdir = tempfile.mkdtemp()

        module = Clustering(
            model=jmvae_model,
            output=tmpdir,
            test_dataset=dataset,
            train_dataset=dataset,
        )

        output = module.eval()
        assert isinstance(output, ModelOutput)
        assert hasattr(output, "cluster_accuracy")
        assert output.cluster_accuracy >= 0 and output.cluster_accuracy <= 1



from pydantic.dataclasses import dataclass
from pythae.models.base.base_config import BaseConfig


@dataclass
class EvaluatorConfig(BaseConfig):
    """

    Base config class for the evaluation modules.

    Args:
        batch_size (int) : The batch size to use in the evaluation. Default to 512
        wandb_path (str) : The user can provide the path of the wandb run with a
            format 'entity/projet_name/run_id' where the metrics should be logged.
            For an existing run (the training run), the info can be found in the training dir
                (in wandb_info.json)
            at the end of training (if wandb was used) or on the hugging_face webpage of the run.
            Otherwise the user can create a new wandb run and get the path with :
                `
                import wandb
                run = wandb.init(entity = your_entity, project=your_project)
                wandb_path = run.path

            If None are provided, the metrics are not logged on wandb.
            Default to None.

    """

    batch_size: int = 512
    wandb_path: str = None



from pydantic.dataclasses import dataclass

from ..joint_models import BaseJointModelConfig


@dataclass
class JMVAEConfig(BaseJointModelConfig):
    """
    This is the base config for the JMVAE model.

    Args:
        n_modalities (int): The number of modalities. Default: None.
        latent_dim (int): The dimension of the latent space. Default: None.
        input_dims (dict[str,tuple]) : The modalities'names (str) and input shapes (tuple).
        uses_likelihood_rescaling (bool): To mitigate modality collapse, it is possible to use likelihood rescaling.
            (see : https://proceedings.mlr.press/v162/javaloy22a.html).
            The inputs_dim must be provided to compute the likelihoods rescalings. It is used in a number of models
            which is why we include it here. Default to False.
        decoders_dist (Dict[str, Union[function, str]]). The reconstruction loss to use per modality.
            Per modality, you can provide a string in ['normal','bernoulli','laplace']. If None is provided, a normal distribution is used for each modality.
        alpha (float):  the parameter that controls the tradeoff between the ELBO and the
            regularization term. Default to 0.1.
        warmup (int): The number of warmup epochs during training. The JMVAE model uses annealing.
            The KL terms in the objective are weighted by a factor beta which is linearly brought to
            1 during the first warmup epochs. Default to 10.
        use_default_joint (bool) :  A boolean encoding if the joint encoder used is the default one.

    """

    alpha: float = 0.1
    warmup: int = 10



from pydantic.dataclasses import dataclass

from ..joint_models import BaseJointModelConfig


@dataclass
class TELBOConfig(BaseJointModelConfig):
    """
    Configuration class for the TELBO model from (arXiv:1705.10762 [cs, stat])
    "Generative Models of Visually Grounded Imagination" (Vedantam et al,2018).


    Args:
        n_modalities (int): The number of modalities. Default: None.
        latent_dim (int): The dimension of the latent space. Default: None.
        input_dims (dict[str,tuple]) : The modalities'names (str) and input shapes (tuple).
        uses_likelihood_rescaling (bool): To mitigate modality collapse, it is possible to use likelihood rescaling.
            (see : https://proceedings.mlr.press/v162/javaloy22a.html).
            The inputs_dim must be provided to compute the likelihoods rescalings. It is used in a number of models
            which is why we include it here. Default to False.
        decoders_dist (Dict[str, Union[function, str]]). The decoder distributions to use per modality.
            Per modality, you can provide a string in ['normal','bernoulli','laplace']. If None is provided,
            a normal distribution is used for each modality.
        decoder_dist_params (Dict[str,dict]) : Parameters for the output decoder distributions, for
            computing the log-probability.
            For instance, with normal or laplace distribution, you can pass the scale in this dictionary.
            ex :  {'mod1' : {scale : 0.75}}
        warmup (int): How many epochs to train the joint encoder and decoders before freezing them
            and learn the unimodal encoders. It is recommended to use half of the
            total training time for the first step. Default to 10.
        lambda_factors (dict[str,float]) : Ponderation factors for the reconstructions in the Joint Elbo.
            If None is provided but uses_likelihood_rescaling is True, we use the inverse product of
            dimensions as a rescaling factor for each modality. If None is provided and uses_likelihood_rescaling
            is False, each factor is set to one. Default to None.
        gamma_factors (dict[str,float]) : Ponderation factors for the reconstructions in the unimodal elbos.
            If None is provided but uses_likelihood_rescaling is True, we use the inverse product of
            dimensions as a rescaling factor for each modality. If None is provided and uses_likelihood_rescaling
            is False, each factor is set to one. Default to None.
        uses_likelihood_rescaling (bool) : Indicates how to set lambda or gamma factors when None are provided.
            Ignored when lambda_factors and gamma_factors are provided. Default to True.
    """

    warmup: int = 10
    lambda_factors: dict = None
    gamma_factors: dict = None
    uses_likelihood_rescaling: bool = True



from typing import Dict

from pythae.models.base.base_utils import ModelOutput
from pythae.models.nn import BaseEncoder
from torch.nn import Module, ModuleDict

from ...data.datasets import MultimodalBaseDataset
from .dcca_config import DCCAConfig
from .objectives import cca_loss, mcca_loss


class DCCA(Module):
    def __init__(self, config: DCCAConfig, networks: Dict[str, BaseEncoder]) -> None:
        super().__init__()

        self.config = config
        self.n_modalities = config.n_modalities
        self.latent_dim = config.embedding_dim
        self.use_all_singular_values = config.use_all_singular_values
        self.set_networks(networks)

        if self.config.n_modalities == 2:
            self.loss = cca_loss(self.latent_dim, self.use_all_singular_values).loss
        else:
            self.loss = mcca_loss(self.latent_dim, self.use_all_singular_values).loss

    def set_networks(self, networks):
        self.networks = ModuleDict()
        assert (
            len(networks) == self.n_modalities
        ), "The number of provided networks doesn't match the number of modalities."

        for m in networks:
            if not isinstance(networks[m], BaseEncoder):
                raise AttributeError(
                    "The DCCA networks must be instances of pythae BaseEncoder class."
                )
            if networks[m].latent_dim != self.latent_dim:
                raise AttributeError(
                    "The DCCA networks must have the same latent dim as the DCCA model"
                    f" itself. ({networks[m].latent_dim} is different {self.latent_dim})"
                )
            self.networks[m] = networks[m]

    def forward(self, inputs: MultimodalBaseDataset):
        if not all([k in self.networks for k in inputs.data]):
            raise AttributeError(
                f"Some inputs keys {inputs.data.keys()} are not known modalities."
                f"The DCCA networks have modalities {self.networks.keys()}"
            )

        embeddings = []
        for m in inputs.data:
            embeddings.append(self.networks[m](inputs.data[m]).embedding)

        # Compute CCA loss or MultiCCA loss between the embeddings

        loss = self.loss(embeddings)

        output = ModelOutput(loss=loss, metrics={})
        return output



from setuptools import find_packages, setup

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(
    name="multivae",
    version="0.1.1",
    author="Agathe Senellart, Clement Chadebec",
    author_email="agathe.senellart@inria.fr",
    description="Unifying Generative Multimodel Variational Autoencoders in Pytorch",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/AgatheSenellart/MultiVae",
    project_urls={
        "Bug Tracker": "https://github.com/AgatheSenellart/MultiVae/issues"
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
    ],
    package_dir={"": "src"},
    packages=find_packages(where="src"),
    install_requires=[
        "pythae",
        "cloudpickle>=2.1.0",
        "imageio",
        "numpy>=1.19",
        "pydantic>=1.8.2",
        "scikit-learn",
        "scipy>=1.7.1",
        "torch>=1.10.1",
        "tqdm",
        "typing_extensions",
        "dataclasses>=0.6",
        "torchvision",
        "pandas",
        "nltk"
    ],
    python_requires=">=3.8",
)



"""
In this section, you will find all the `built-in` datasets that are currently implemented in `multivae` library
"""

from .base import DatasetOutput, IncompleteDataset, MultimodalBaseDataset
from .cub import CUB
from .mnist_svhn import MnistSvhn

__all__ = ["MultimodalBaseDataset", "MnistSvhn", "IncompleteDataset", "DatasetOutput", "CUB"]





import os

import numpy as np
import torch
from PIL import Image
from torch.utils.data import DataLoader
from torchvision.utils import make_grid

from multivae.data import MultimodalBaseDataset
from multivae.data.datasets.utils import adapt_shape
from multivae.data.utils import set_inputs_to_device
from multivae.metrics.base.evaluator_config import EvaluatorConfig
from multivae.models.base import BaseMultiVAE, ModelOutput
from multivae.samplers.base import BaseSampler

from ..base.evaluator_class import Evaluator
from .clustering_config import ClusteringConfig


class Clustering(Evaluator):

    """
    Module to perform clustering in the latent space.
    As of now, it is only supported for the joint representation of the data.
    The eval() function fits a k-means model on the training embeddings, then uses this model
    to classify the test_samples and returns a k-means accuracy for this prediction.

    Args:
        model (BaseMultiVAE) : The model to evaluate.
        test_dataset (MultimodalBaseDataset) : The dataset to use for computing the metrics.
        train_dataset (MultimodalBaseDataset): The training dataset to fit the k-means.
        output (str) : The folder path to save metrics. The metrics will be saved in a metrics.txt
            file.
        eval_config (EvaluatorConfig) : The configuration class to specify parameters for the
            evaluation.
        sampler (BaseSampler) : A custom sampler for sampling from the common latent space. If
            None is provided, samples are generated from the prior. Not used in this module, just
            given to keep the API for evaluators modules.

    """

    def __init__(
        self,
        model: BaseMultiVAE,
        test_dataset: MultimodalBaseDataset,
        train_dataset: MultimodalBaseDataset,
        output: str = None,
        eval_config=ClusteringConfig(),
        sampler: BaseSampler = None,
    ) -> None:
        super().__init__(model, test_dataset, output, eval_config, sampler)

        self.num_samples_for_fit = eval_config.num_samples_for_fit
        self.n_fits = eval_config.number_of_runs
        self.use_mean = eval_config.use_mean
        if eval_config.clustering_method == "kmeans":
            try:
                from sklearn.cluster import KMeans
            except:
                raise ModuleNotFoundError(
                    "scikit-learn must be installed to perform clustering. Run `pip install scikit-learn` to install it "
                )
            self.clustering = KMeans(n_clusters=eval_config.n_clusters, max_iter=300)

            self.train_dataset = train_dataset

    def fit_clustering(self, mods="all"):
        # compute all training embeddings
        dl = DataLoader(self.train_dataset, self.batch_size, shuffle=True)

        list_z = []
        n_samples = 0
        labels = []
        for inputs in dl:
            if (
                self.num_samples_for_fit is not None
                and n_samples > self.num_samples_for_fit
            ):
                break
            inputs = set_inputs_to_device(inputs, self.device)
            with torch.no_grad():
                list_z.append(
                    self.model.encode(inputs, mods, return_mean=self.use_mean).z
                )
            if inputs.labels is not None:
                labels.append(inputs.labels)

        if len(labels) > 0:
            labels = torch.cat(labels).cpu().numpy()
            labels.dtype = np.int64
        all_z = torch.cat(list_z).cpu().numpy()
        clusters_labels = self.clustering.fit_predict(all_z)
        # Get the majority label for each cluster
        self.labels_dict = {str(m): m for m in np.unique(clusters_labels)}
        if len(labels) == len(clusters_labels):
            for c in np.unique(clusters_labels):
                maj_value = np.bincount(labels[clusters_labels == c]).argmax()
                self.labels_dict[str(c)] = maj_value

    def cluster_accuracy(self, mods="all"):
        mean_acc = []
        for i in range(self.n_fits):
            self.fit_clustering(mods)

            # Cluster the test dataset
            acc = 0
            n_samples = 0
            for inputs in self.test_loader:
                inputs = set_inputs_to_device(inputs, self.device)
                with torch.no_grad():
                    z = self.model.encode(inputs, mods, return_mean=self.use_mean).z
                clabels = self.clustering.predict(z.cpu().numpy())
                labels = np.array([self.labels_dict[str(c)] for c in clabels])
                true_labels = inputs.labels.cpu().numpy()
                acc += np.sum(labels == true_labels)
                n_samples += len(z)

            accuracy = acc / n_samples
            mean_acc.append(accuracy)
        accuracy = np.mean(mean_acc)
        self.metrics["cluster_accuracy"] = accuracy
        self.logger.info(f"Cluster accuracy is {accuracy}")
        return ModelOutput(cluster_accuracy=accuracy)

    def eval(self):
        output = self.cluster_accuracy("all")
        self.log_to_wandb()
        return output



from typing import Dict, List, Union

from pydantic.dataclasses import dataclass

from ..base.base_config import BaseMultiVAEConfig


@dataclass
class MoPoEConfig(BaseMultiVAEConfig):
    """
    This class is the configuration class for the MoPoE model, from
    'Generalized Multimodal ELBO' Sutter 2021 (https://arxiv.org/abs/2105.02470)


    Args:
        n_modalities (int): The number of modalities. Default: None.
        latent_dim (int): The dimension of the latent space. Default: None.
        input_dims (dict[str,tuple]) : The modalities'names (str) and input shapes (tuple).
        uses_likelihood_rescaling (bool): To mitigate modality collapse, it is possible to use likelihood rescaling.
            (see : https://proceedings.mlr.press/v162/javaloy22a.html).
            The inputs_dim must be provided to compute the likelihoods rescalings. Default to True.
        decoders_dist (Dict[str, Union[function, str]]). The decoder distributions to use per modality.
            Per modality, you can provide a string in ['normal','bernoulli','laplace']. If None is provided,
            a normal distribution is used for each modality.
        decoder_dist_params (Dict[str,dict]) : Parameters for the output decoder distributions, for
            computing the log-probability.
            For instance, with normal or laplace distribution, you can pass the scale in this dictionary.
            ex :  {'mod1' : {scale : 0.75}}
        subsets (Union[List[list], Dict[list]]) : List containing the subsets to consider. If None is provided,
            all subsets are considered. Example of valid input : [['mod_1', 'mod_2'], ['mod_1'], ['mod_2']].
            Default to None.
        beta (float) : The weight to the KL divergence term in the ELBO. Default to 1.0
        use_modality_specific_spaces (bool) : Whether to use additional latent spaces for modalities. Default to False.
        beta_style (float) : The beta factor for additional elbos in the case, there are multiple latent spaces.
            Default to 1.
        modalities_specific_dim (dict) : a dictionary containing the modalities names and the dimension of
            the additional latent space for each modality. Default to None.

    """

    subsets: Union[List[list], Dict[str, list]] = None
    beta: float = 1.0
    use_modality_specific_spaces: bool = False
    beta_style: float = 1.0
    modalities_specific_dim: dict = None
